#  ----------------------------------------------------------------------------
#  "THE BEER-WARE LICENSE" (Revision 42):
#  dkratzert@gmx.de> wrote this file.  As long as you retain
#  this notice you can do whatever you want with this stuff. If we meet some day,
#  and you think this stuff is worth it, you can buy me a beer in return.
#  Dr. Daniel Kratzert
#  ----------------------------------------------------------------------------
import os
import re
from collections import namedtuple
from collections.abc import Generator
from contextlib import suppress
from pathlib import Path
from typing import Any, Generator

import gemmi

from fastmolwidget.atoms import atoms
from fastmolwidget.tools import grouper, isnumeric

if hasattr(gemmi, 'set_leak_warnings'):
    gemmi.set_leak_warnings(False)

from gemmi.cif import as_string, Document, Loop, quote

adp = namedtuple('adp', ('label', 'U11', 'U22', 'U33', 'U23', 'U13', 'U12'))


class GemmiError(Exception):
    pass


class CifReader:
    """
    This class holds the content of a cif file, independent of the file parser used.
    """

    def __init__(self, file: Path | str, new_block: str = ''):
        """

        Args:
            file: CIF file to open
            new_block: Create a new block (new file) if a name is given. Otherwise, just
                       the existing document is opened.
        """
        self.chars_ok = True
        self._abs_hkl_details: dict[str, str] = {}
        if isinstance(file, str):
            self.fileobj = Path(file)
        elif isinstance(file, Path):
            self.fileobj = file
        else:
            raise TypeError('The file parameter must be string or Path object.')
        self.current_block = ''
        # I do this in small steps instead of gemmi.cif.read_file() in order to
        # leave out the check_for_missing_values. This way, gemmi reads cif files even
        # with missing values.
        if new_block:
            self.doc: Document = gemmi.cif.Document()
            self.doc.add_new_block(new_block)
        else:
            try:
                self.doc = self.read_file(self.fileobj)
            except Exception as e:
                raise
                raise GemmiError(e)
        # Starting with first block, but can use others with subsequent self._onload():
        self.block: gemmi.cif.Block = self.doc[0]
        self._on_load()

    @property
    def path_base(self) -> Path:
        """
        The absolute path of the current file without file name:
        /foo/bar/baz
        """
        return Path(self.doc.source).resolve().parent

    @property
    def filename(self) -> str:
        """
        The name of the current file without path:
        foo.cif
        """
        return Path(self.doc.source).name or self.fileobj.name

    def get_line_numbers_of_bad_characters(self, filepath: Path):
        line_numbers = []
        for num, line in enumerate(filepath.read_bytes().splitlines(keepends=True)):
            try:
                line.decode('ascii')
            except UnicodeDecodeError:
                line_numbers.append(num)
        return line_numbers

    @property
    def is_multi_cif(self) -> bool:
        return True if len(self.doc) > 1 else False

    def load_this_block(self, index: int) -> None:
        self.block = self.doc[index]
        self.current_block = self.block.name
        self._on_load()

    def load_block_by_name(self, blockname: str) -> None:
        self.block = self.doc.find_block(blockname)
        self.current_block = self.block.name
        self._on_load()

    def delete_block(self, index: int) -> None:
        del self.doc[index]
        self.block: gemmi.cif.Block = self.doc[0]

    def _on_load(self) -> None:
        # will not ok with non-ascii characters in the res file:
        self.chars_ok = True
        try:
            self.doc.check_for_duplicates()
        except RuntimeError as e:
            raise GemmiError(e)
        self.atomic_struct: gemmi.SmallStructure = gemmi.make_small_structure_from_block(self.block)
        # A dictionary to convert Atom names like 'C1_2' or 'Ga3' into Element names like 'C' or 'Ga'
        self._name2elements = dict(zip([x.upper() for x in self.block.find_loop('_atom_site_label')],
                                       [y.upper() for y in self.block.find_loop('_atom_site_type_symbol')],
                                       strict=False))

    def read_file(self, path: Path) -> gemmi.cif.Document:
        """
        Reads a cif file and returns a gemmi document object.
        """
        doc = gemmi.cif.Document()
        doc.source = str(path.resolve())
        doc.parse_file(str(path.resolve()))
        return doc

    def read_string(self, cif_string: str) -> gemmi.cif.Document:
        """
        Reads a cif file from a string and returns a gemmi cif docment.
        :param cif_string: cif as string
        :return: gemmi document
        """
        doc = gemmi.cif.Document()
        doc.parse_string(cif_string)
        return doc

    def cif_as_string(self, without_hkl=False) -> str:
        opt = gemmi.cif.WriteOptions()
        opt.align_pairs = 33
        if without_hkl:
            # return a copy, do not delete hkl from original:
            doc = gemmi.cif.Document()
            doc.parse_string(self.doc.as_string(options=opt))
            for block in doc:
                if block.find_pair_item('_shelx_hkl_file'):
                    block.find_pair_item('_shelx_hkl_file').erase()
                if block.find_pair_item('_shelx_fcf_file'):
                    block.find_pair_item('_shelx_fcf_file').erase()
            return doc.as_string(options=opt)
        else:
            return self.doc.as_string(options=opt)

    def __getitem__(self, item: str) -> str:
        """
        This method returns an empty string when the item value is '?'
        """
        if value := self.block.find_value(item):
            return as_string(value)
        else:
            return ''

    def __setitem__(self, key: str, value: str) -> None:
        """Set a key value pair of the current block.
        Values are automatically encoded from utf-8 and delimited.
        """
        self.set_pair_delimited(key, value)

    def __delitem__(self, key: str) -> None:
        with suppress(AttributeError):
            self.block.find_pair_item(key).erase()

    def __contains__(self, item) -> bool:
        return bool(self.__getitem__(item))

    def __str__(self) -> str:
        return (f"CIF file: {self.fileobj.resolve()!s}\n"
                f"{len(self.doc)} Block(s): {', '.join([x.name for x in self.doc])}\n"
                f"Contains SHELX res file: {True if self.res_file_data else False}\n"
                f"Has {self.natoms()} atoms"
                f", {self.nbonds()} bonds"
                f", {self.nangles()} angles")

    def file_is_there_and_writable(self) -> bool:
        return self.fileobj.exists() and self.fileobj.is_file() and os.access(self.fileobj, os.W_OK)

    def set_pair_delimited(self, key: str, txt: str):
        """
        Converts special characters to their markup counterparts.
        """
        # txt = utf8_to_str(txt)
        try:
            # bad hack to get the numbered values correct
            float(txt)
            self.block.set_pair(key, txt)
        except (TypeError, ValueError):
            # prevent _key '?' in cif:
            if txt == '?':
                self.block.set_pair(key, txt)
            else:
                self.block.set_pair(key, quote(txt))

    def save(self, filename: Path | None = None) -> None:
        """
        Saves the current cif file. It uses the '[basename]-finalcif.cif' suffix by default.
        :param filename:  Name to save cif file to if the default is not sufficient.
        """
        if self.is_empty():
            print(f'DBG> CIF file {filename} is empty.')
            return
        if not self.is_writable(filename):
            raise PermissionError(f'Failed to open {filename.resolve()} for writing: Operation not permitted')
        print('Saving CIF to', Path(filename).resolve())
        options = gemmi.cif.WriteOptions()
        options.prefer_pairs = False
        options.compact = False
        options.align_pairs = 33
        options.align_loops = 15
        options.misuse_hash = False
        self.doc.write_file(str(filename), options=options)

    def is_writable(self, filepath: Path | str) -> bool:
        filepath = Path(filepath)
        try:
            exist = filepath.exists()
            with open(filepath, 'wb+') as f:
                f.write(b'')
            if not exist:
                Path(filepath).unlink(missing_ok=True)
            return True
        except Exception:
            return False

    @property
    def res_file_data(self) -> str:
        try:
            if self['_shelx_res_file']:
                return self['_shelx_res_file']
            elif self['_iucr_refine_instructions_details']:
                return self['_iucr_refine_instructions_details']
            else:
                return ''
        except UnicodeDecodeError:
            # This is a fallback in case _shelx_res_file has non-ascii characters.
            print('File has non-ascii characters. Switching to compatible mode.')
            self.doc = self.read_string(self.fileobj.read_text(encoding='latin1', errors='ignore'))
            self.block = self.doc.sole_block()
            self.chars_ok = False
            return self.block.find_value('_shelx_res_file')

    @property
    def hkl_file(self) -> str:
        hkl_loop: Loop = self.get_loop('_diffrn_refln_index_h')
        if hkl_loop and hkl_loop.width() > 4:
            return self._hkl_from_cif_format(hkl_loop)
        else:
            # returns an empty string if no cif hkl was found:
            return self._hkl_from_shelx()

    def _hkl_from_cif_format(self, hkl_loop: Loop) -> str:
        hkl_list = []
        format_string = '{:>4}{:>4}{:>4}{:>8}{:>8}'
        if hkl_loop.width() == 6:
            format_string = '{:>4}{:>4}{:>4}{:>8}{:>8}{:>4}'
        for line in grouper(hkl_loop.values, hkl_loop.width()):
            hkl_list.append(format_string.format(*line[:6]))
        return '\n'.join(hkl_list)

    def _hkl_from_shelx(self) -> str:
        try:
            if self['_shelx_hkl_file']:
                return self['_shelx_hkl_file']
            elif self['_iucr_refine_reflections_details']:
                return self['_iucr_refine_reflections_details']
            elif self['_xd_hkl_file']:
                return self['_xd_hkl_file']
            else:
                return ''
        except Exception as e:
            print(f'No hkl data found in CIF!, {e}')
            return ''

    @property
    def fab_file(self):
        if self['_shelx_fab_file']:
            return self['_shelx_fab_file']
        return ''

    @property
    def hkl_file_without_foot(self) -> str:
        """Returns a hkl file with no content after the 0 0 0 reflection"""
        zero_reflection_position = self._find_line_of_000(self.hkl_file)
        if zero_reflection_position:
            return '\n'.join(self.hkl_file.splitlines(keepends=False)[:zero_reflection_position + 1])
        else:
            return self.hkl_file

    @staticmethod
    def _find_line_of_000(lines: str) -> int:
        pattern = re.compile(r'^\s+0\s+0\s+0\s+0.*')
        for num, line in enumerate(lines.splitlines(keepends=False)):
            found = pattern.match(line)
            if found and num > 0:
                return num
        return 0

    def normal_search(self, hkl) -> int:
        pattern = re.compile(r'\s+0\s+0\s+0\s+0')
        found = pattern.search(hkl)
        if found:
            zero_reflection_position = found.start()
        else:
            zero_reflection_position = 0
        return zero_reflection_position

    def reversed_search(self, hkl_splitted: list[str]) -> int:
        pattern = re.compile(r'\s+0\s+0\s+0\s+0')
        zero_reflection_position = 0
        for num, line in enumerate(reversed(hkl_splitted)):
            found = pattern.search(line)
            if num > 500:
                # A longer footer is not realistic
                break
            if found:
                zero_reflection_position = len(hkl_splitted) - num
        return zero_reflection_position

    def is_empty(self) -> bool:
        """
        Checks if the currebt CIF contains data, but this works only for single-CIFs!
        """
        if len(self.doc) < 2 and len(self.keys()) + len(self.loops) == 0:
            return True
        return False

    def keys(self) -> list[str]:
        """
        Returns a plain list of keys that are really in this CIF.
        """
        return [x.pair[0] for x in self.block if x.pair is not None]

    def values(self):
        """
        Returns a plain list of keys that are really in this CIF.
        """
        return [x.pair[1] for x in self.block if x.pair is not None]

    @property
    def loops(self) -> list[gemmi.cif.Loop]:
        """
        Returns a list of loops contained in the current block.
        """
        loops = []
        for b in self.block:
            if b.loop:
                loops.append(b.loop)
        return loops

    def add_loop_to_cif(self, loop_tags: list[str], row_values: list[str] | tuple[str]) -> gemmi.cif.Loop:
        gemmi_loop = self.init_loop(loop_tags)
        for row in list(grouper(row_values, len(loop_tags))):
            gemmi_loop.add_row(row)
        return gemmi_loop

    def add_loop_from_columns(self, loop_tags: list[str],
                              column_values: list[list[str]] | tuple[tuple[str]]) -> gemmi.cif.Loop:
        try:
            gemmi_loop = self.init_loop(loop_tags)
            gemmi_loop.set_all_values(column_values)
        except RuntimeError as err:
            raise Exception('Gemmi raised an exception') from err
        return gemmi_loop

    @property
    def n_loops(self) -> int:
        return len(self.loops)

    def get_loop(self, key_in_loop: str) -> gemmi.cif.Loop:
        return self.block.find_loop(key_in_loop).get_loop()

    def get_loop_column(self, key_in_loop: str) -> list:
        return [as_string(x) for x in self.block.find_loop(key_in_loop)]

    def init_loop(self, loop_keywords: list) -> gemmi.cif.Loop:
        return self.block.init_loop('', loop_keywords)

    @property
    def z_value(self) -> float:
        return self.atomic_struct.cell.volume / self.atomic_struct.cell.volume_per_image()

    def _spgr(self) -> gemmi.SpaceGroup:
        if self.symmops and self.symmops != ['']:
            symm_ops = self.symmops
        else:
            symm_ops = self.symmops_from_spgr
        return gemmi.find_spacegroup_by_ops(gemmi.GroupOps([gemmi.Op(o) for o in symm_ops]))

    @property
    def space_group(self) -> str:
        """
        Returns the space group from the symmetry operators.
        spgr.short_name() gives the short name.
        """
        try:
            return self._spgr().xhm()
        except (AttributeError, RuntimeError, ValueError):
            if self['_space_group_name_H-M_alt']:
                return gemmi.cif.as_string(self['_space_group_name_H-M_alt'])
            elif self['_symmetry_space_group_name_H-M']:
                return gemmi.cif.as_string(self['_symmetry_space_group_name_H-M'])
            else:
                return ''

    @property
    def symmops_from_spgr(self) -> list[str]:
        # _symmetry_space_group_name_Hall
        space_group = None
        if self['_space_group_name_H-M_alt']:
            space_group = self['_space_group_name_H-M_alt']
        if self['_symmetry_space_group_name_H-M']:
            space_group = self['_symmetry_space_group_name_H-M']
        if not space_group:
            return []
        ops = [op.triplet() for op in
               gemmi.find_spacegroup_by_name(gemmi.cif.as_string(space_group)).operations()]
        return ops

    @property
    def spgr_number_from_symmops(self) -> int:
        return self._spgr().number

    @property
    def spgr_number(self) -> int:
        if self['_space_group_IT_number'] and isnumeric(self['_space_group_IT_number']):
            return int(self['_space_group_IT_number'])
        elif self['_symmetry_Int_Tables_number'] and isnumeric(self['_symmetry_Int_Tables_number']):
            return int(self['_symmetry_Int_Tables_number'])
        else:
            return self.spgr_number_from_symmops

    @property
    def crystal_system(self) -> str:
        if not self._spgr():
            return ''
        return self._spgr().crystal_system_str()

    @property
    def hall_symbol(self) -> str:
        return self._spgr().hall

    @property
    def hkl_checksum_calcd(self) -> int:
        """
        Calculates the shelx checksum for the hkl file content of a cif file.
        """
        if self.hkl_file:
            return self.calc_checksum(self.hkl_file)
        else:
            return 0

    @property
    def res_checksum_calcd(self) -> int:
        """
        Calculates the shelx checksum for the res file content of a cif file.
        """
        if self.res_file_data:
            return self.calc_checksum(self.res_file_data)
        return 0

    @staticmethod
    def calc_checksum(input_str: str) -> int:
        """
        Calculates the shelx checksum of a cif file.
        The original algorithm was posted by Berthold Stöger on the Bruker Users Mailing list.
        """
        crc_sum = 0
        try:
            bytes_str = input_str.encode('cp1250', 'ignore')
        except Exception:
            bytes_str = input_str.encode('ascii', 'ignore')
        for char in bytes_str:
            # print(char)
            if char > 32:  # ascii 32 is space character
                crc_sum += char
        crc_sum %= 714025
        crc_sum = crc_sum * 1366 + 150889
        crc_sum %= 714025
        crc_sum %= 100000
        return crc_sum

    def rename_data_name(self, newname: str = ''):
        """
        Renames data_ tags to the newname. Also _vrf tags are renamed accordingly.
        http://journals.iucr.org/services/cif/checking/checkfaq.html
        """
        newname = ''.join([i for i in newname if i.isascii()])
        self.block.name = newname
        for item in self.block:
            if item.pair is not None:
                key, value = item.pair
                if key.startswith('_vrf'):
                    newkey = '_' + '_'.join(key.split('_')[1:3]) + '_' + newname
                    self.block.find_pair_item(key).erase()
                    self.block.set_pair(newkey, value)

    @property
    def symmops(self) -> list[str]:
        """
        Reads the symmops from the cif file.
        """
        xyz1 = self.block.find(["_symmetry_equiv_pos_as_xyz"])  # deprecated
        xyz2 = self.block.find(["_space_group_symop_operation_xyz"])  # New definition
        if xyz1:
            return [i.str(0) for i in xyz1]
        elif xyz2:
            return [i.str(0) for i in xyz2]
        else:
            return self.symmops_from_spgr

    @property
    def is_centrosymm(self) -> bool:
        """
        Whether a structure is centrosymmetric or not.
        """
        if not self.symmops or self.symmops == ['']:
            # Do not crash without symmops
            return False
        ops = gemmi.GroupOps([gemmi.Op(o) for o in self.symmops])
        return ops.is_centrosymmetric()

    def atoms(self, without_h: bool = False) -> Generator:
        """
        Atoms from the CIF where values are returned as string like in the CIF with esds.
        """
        labels = self.block.find_loop('_atom_site_label')
        types = self.block.find_loop('_atom_site_type_symbol') or [atoms.get_atomlabel(x) for x in labels]
        x = self.block.find_loop('_atom_site_fract_x')
        y = self.block.find_loop('_atom_site_fract_y')
        z = self.block.find_loop('_atom_site_fract_z')
        part = self.block.find_loop('_atom_site_disorder_group') or ('0',) * len(labels)
        occ = self.block.find_loop('_atom_site_occupancy') or ('1.000000',) * len(labels)
        u_eq = self.block.find_loop('_atom_site_U_iso_or_equiv') or ('.',) * len(labels)
        atom = namedtuple('atom', ('label', 'type', 'x', 'y', 'z', 'part', 'occ', 'u_eq'))
        for label, type, x, y, z, part, occ, u_eq in zip(labels, types, x, y, z,
                                                         part,
                                                         occ,
                                                         u_eq, strict=False):
            if without_h and self.ishydrogen(label):
                continue
            #         0    1   2  3  4   5   6     7
            # yield label, type, x, y, z, part, occ, ueq
            yield atom(label=label, type=type, x=x, y=y, z=z, part=part, occ=occ, u_eq=u_eq)

    @property
    def atoms_fract(self) -> Generator:
        """Atoms with numbers as float values without esd."""
        for at in self.atomic_struct.sites:
            yield [at.label, at.type_symbol, at.fract.x, at.fract.y, at.fract.z, at.disorder_group, at.occ, at.u_iso]

    @property
    def atoms_orth(self) -> Generator:
        atom = namedtuple('atom', ('label', 'type', 'x', 'y', 'z', 'part', 'occ', 'u_eq'))
        cell = self.atomic_struct.cell
        for at in self.atomic_struct.sites:
            x, y, z = at.orth(cell)
            yield atom(label=at.label, type=at.type_symbol, x=x, y=y, z=z,
                       part=at.disorder_group, occ=at.occ, u_eq=at.u_iso)

    def displacement_parameters(self) -> Generator[adp, Any, None]:
        """
        Yields the anisotropic displacement parameters.
        """
        labels = self.block.find_loop('_atom_site_aniso_label')
        u11 = self.block.find_loop('_atom_site_aniso_U_11')
        u22 = self.block.find_loop('_atom_site_aniso_U_22')
        u33 = self.block.find_loop('_atom_site_aniso_U_33')
        u23 = self.block.find_loop('_atom_site_aniso_U_23')
        u13 = self.block.find_loop('_atom_site_aniso_U_13')
        u12 = self.block.find_loop('_atom_site_aniso_U_12')
        for label, u11, u22, u33, u23, u13, u12 in zip(labels, u11, u22, u33, u23, u13, u12, strict=True):
            yield adp(label=label, U11=u11, U22=u22, U33=u33, U12=u12, U13=u13, U23=u23)

    @property
    def hydrogen_atoms_present(self) -> bool:
        for at in self.atomic_struct.sites:
            if at.type_symbol in ('H', 'D'):
                return True
        return False

    @property
    def disorder_present(self) -> bool:
        for at in self.atomic_struct.sites:
            if at.disorder_group:
                return True
        return False

    @property
    def cell(self):
        c = self.atomic_struct.cell
        cell = namedtuple('cell', 'a, b, c, alpha, beta, gamma, volume')
        return cell(c.a, c.b, c.c, c.alpha, c.beta, c.gamma, c.volume)

    def ishydrogen(self, label: str) -> bool:
        """
        Determines if an atom with the name 'label' is a hydrogen atom.
        """
        hydrogen = ('H', 'D')
        if self._iselement(label) in hydrogen:
            return True
        else:
            return False

    def checksymm(self, symm):
        """Add translation of 555 to symmetry elements without translation"""
        if len(symm) == 1 and symm not in {'.', '?'}:
            symm = symm + '_555'
        return symm

    def bonds(self, without_h: bool = False) -> Generator:
        """
        Yields a list of bonds in the cif file.
        """
        label1 = self.block.find_loop('_geom_bond_atom_site_label_1')
        label2 = self.block.find_loop('_geom_bond_atom_site_label_2')
        dist = self.block.find_loop('_geom_bond_distance')
        symm = self.block.find_loop('_geom_bond_site_symmetry_2')
        publ_loop = self.block.find_loop('_geom_bond_publ_flag') or len(label1) * ['?']
        bond = namedtuple('bond', ('label1', 'label2', 'dist', 'symm'))
        for label1, label2, dist, symm, publ in zip(label1, label2, dist, symm, publ_loop, strict=False):
            if (without_h and (self.ishydrogen(label1) or self.ishydrogen(label2))) or self.yes_not_set(publ):
                continue
            else:
                yield bond(label1=label1, label2=label2, dist=dist, symm=self.checksymm(symm))

    def bond_dist(self, pair: str) -> str | None:
        for p in self.bonds():
            if f'{p.label1}-{p.label2}' == pair:
                return p.dist
        return None

    def angle(self, atoms: str) -> str | None:
        for p in self.angles():
            if f'{p.label1}-{p.label2}-{p.label3}' == atoms:
                return p.angle_val
        return None

    def torsion(self, atoms: str) -> str | None:
        for p in self.torsion_angles():
            if f'{p.label1}-{p.label2}-{p.label3}-{p.label4}' == atoms:
                return p.torsang
        return None

    def angles(self, without_H: bool = False) -> Generator:
        label1 = self.block.find_loop('_geom_angle_atom_site_label_1')
        label2 = self.block.find_loop('_geom_angle_atom_site_label_2')
        label3 = self.block.find_loop('_geom_angle_atom_site_label_3')
        angle_val = self.block.find_loop('_geom_angle')
        symm1 = self.block.find_loop('_geom_angle_site_symmetry_1')
        symm2 = self.block.find_loop('_geom_angle_site_symmetry_3')
        publ_loop = self.block.find_loop('_geom_angle_publ_flag') or len(label1) * ['?']
        angle = namedtuple('angle', ('label1', 'label2', 'label3', 'angle_val', 'symm1', 'symm2'))
        for label1, label2, label3, angle_val, symm1, symm2, publ in (
            zip(label1, label2, label3, angle_val, symm1, symm2, publ_loop, strict=False)):
            if ((without_H and (self.ishydrogen(label1) or self.ishydrogen(label2) or self.ishydrogen(label3))) or
                self.yes_not_set(publ)):
                continue
            else:
                yield angle(label1=label1, label2=label2, label3=label3, angle_val=angle_val,
                            symm1=self.checksymm(symm1), symm2=self.checksymm(symm2))

    def _iselement(self, name: str) -> str:
        try:
            return self._name2elements[name.upper()]
        except KeyError:
            return atoms.get_atomlabel(name)

    def natoms(self, without_h: bool = False) -> int:
        return len(tuple(self.atoms(without_h)))

    def nbonds(self, without_h: bool = False) -> int:
        """
        Number of bonds in the cif object, with and without hydrogen atoms.
        """
        return len(tuple(self.bonds(without_h)))

    def nangles(self, without_h: bool = False) -> int:
        """
        Number of bond angles in the cif object, with and without hydrogen atoms.
        """
        return len(tuple(self.angles(without_h)))

    def ntorsion_angles(self, without_h: bool = False) -> int:
        """
        Number of torsion angles in the cif object, with and without hydrogen atoms.
        """
        return len(tuple(self.torsion_angles(without_h)))

    def torsion_angles(self, without_h: bool = False) -> Generator:
        label1 = self.block.find_loop('_geom_torsion_atom_site_label_1')
        label2 = self.block.find_loop('_geom_torsion_atom_site_label_2')
        label3 = self.block.find_loop('_geom_torsion_atom_site_label_3')
        label4 = self.block.find_loop('_geom_torsion_atom_site_label_4')
        torsang = self.block.find_loop('_geom_torsion')
        symm1 = self.block.find_loop('_geom_torsion_site_symmetry_1')
        symm2 = self.block.find_loop('_geom_torsion_site_symmetry_2')
        symm3 = self.block.find_loop('_geom_torsion_site_symmetry_3')
        symm4 = self.block.find_loop('_geom_torsion_site_symmetry_4')
        publ_loop = self.block.find_loop('_geom_torsion_publ_flag') or len(label1) * ['?']
        tors = namedtuple('Torsion',
                          ('label1', 'label2', 'label3', 'label4', 'torsang', 'symm1', 'symm2', 'symm3', 'symm4'))
        for label1, label2, label3, label4, torsang, symm1, symm2, symm3, symm4, publ in zip(label1, label2, label3,
                                                                                             label4,
                                                                                             torsang, symm1, symm2,
                                                                                             symm3,
                                                                                             symm4, publ_loop,
                                                                                             strict=False):
            if ((without_h and (self.ishydrogen(label1) or self.ishydrogen(label2)
                                or self.ishydrogen(label3) or self.ishydrogen(label3))) or self.yes_not_set(publ)):
                continue
            yield tors(label1=label1, label2=label2, label3=label3, label4=label4, torsang=torsang,
                       symm1=self.checksymm(symm1),
                       symm2=self.checksymm(symm2),
                       symm3=self.checksymm(symm3),
                       symm4=self.checksymm(symm4))

    def hydrogen_bonds(self) -> Generator:
        label_d = self.block.find_loop('_geom_hbond_atom_site_label_D')
        label_h = self.block.find_loop('_geom_hbond_atom_site_label_H')
        label_a = self.block.find_loop('_geom_hbond_atom_site_label_A')
        dist_dh = self.block.find_loop('_geom_hbond_distance_DH')
        dist_ha = self.block.find_loop('_geom_hbond_distance_HA')
        dist_da = self.block.find_loop('_geom_hbond_distance_DA')
        angle_dha = self.block.find_loop('_geom_hbond_angle_DHA')
        symm = self.block.find_loop('_geom_hbond_site_symmetry_A')
        publ_loop = self.block.find_loop('_geom_hbond_publ_flag') or len(label_d) * ['?']
        hydr = namedtuple('HydrogenBond', ('label_d', 'label_h', 'label_a', 'dist_dh', 'dist_ha', 'dist_da',
                                           'angle_dha', 'symm'))
        for label_d, label_h, label_a, dist_dh, dist_ha, dist_da, angle_dha, symm, publ in (
            zip(label_d, label_h, label_a, dist_dh, dist_ha, dist_da, angle_dha, symm, publ_loop, strict=False)):
            if self.yes_not_set(publ):
                continue
            yield hydr(label_d, label_h, label_a, dist_dh, dist_ha, dist_da, angle_dha, self.checksymm(symm))

    def yes_not_set(self, publ: str):
        return publ not in {'y', 'yes', '?', '.'} or publ in {'n', 'no'}

    def key_value_pairs(self) -> list[tuple[str, str]]:
        """
        Returns the key/value pairs of a cif file sorted by priority.
        """
        keys_without_values, keys_with_values = self.keys_with_essentials()
        return [*keys_without_values, ('These below are already in:', '---------------------'), *keys_with_values]

    def pairs(self):
        for item in self.block:
            if item.pair is not None:
                yield item.pair

    def _is_centrokey(self, key) -> bool:
        """
        Is True if the kurrent key is only valid
        for non-centrosymmetric structures
        """
        return self.is_centrosymm and key in non_centrosymm_keys

    def test_res_checksum(self) -> bool:
        """
        A method to check whether the checksums in the cif file fit to the content.
        """
        cif_res_ckecksum = 0
        res_checksum_calcd = self.res_checksum_calcd
        if res_checksum_calcd > 0:
            cif_res_ckecksum = self['_shelx_res_checksum'] or -1
            try:
                cif_res_ckecksum = int(cif_res_ckecksum)
            except ValueError:
                cif_res_ckecksum = -1
        if cif_res_ckecksum > 0 and cif_res_ckecksum != res_checksum_calcd:
            return False
        else:
            return True

    def test_hkl_checksum(self) -> bool:
        """
        A method to check whether the checksums in the cif file fit to the content.
        """
        cif_hkl_ckecksum = 0
        hkl_checksum_calcd = self.hkl_checksum_calcd
        if hkl_checksum_calcd > 0:
            cif_hkl_ckecksum = self['_shelx_hkl_checksum'] or -1
            try:
                cif_hkl_ckecksum = int(cif_hkl_ckecksum)
            except ValueError:
                cif_hkl_ckecksum = -1
        if cif_hkl_ckecksum > 0 and cif_hkl_ckecksum != hkl_checksum_calcd:
            return False
        else:
            return True


if __name__ == '__main__':
    # c = CifContainer('../41467_2015.cif')
    # c = CifContainer('test-data/p21c.cif')

    c = CifContainer('test-data/DK_Zucker2_0m.cif')
    c.load_this_block(len(c.doc) - 1)
    # pp(list(c.torsion_angles()))
    print(c.bond_dist('C1-C2'))
    # pp(c.hkl_extra_info)
    # print(c.hkl_file)
    # print(c.hkl_as_cif)
    # print(c.test_hkl_checksum())
    # s = Shelxfile()
    # print(c.res_file_data)
    # s.read_string(c.res_file_data)
    # print(s.hklf.n)

    # print(CifContainer('tests/examples/1979688.cif').hkl_as_cif[-250:])

    """
    import n#umpy as np
    mtz = gemmi.Mtz(with_base=True)
    data = [x.split() for x in c.hkl_file.splitlines()]
    mtz.add_column('I', 'I')
    mtz.add_column('SIGI', 'Q')
    mtz.set_data(data)
    mtz.set_cell_for_all(gemmi.UnitCell(c.cell.a, c.cell.b, c.cell.c, c.cell.alpha, c.cell.beta, c.cell.gamma))
    mtz.spacegroup = gemmi.find_spacegroup_by_name(c.space_group)
    mtz.update_reso()
    size = mtz.get_size_for_hkl()
    np.seterr(divide='ignore', invalid='ignore')
    ios =  intensity.array / sigma.array
    
    doc = gemmi.cif.read('tests/examples/1979688-finalcif.fcf')
    rblock = gemmi.hkl_cif_as_refln_block(doc[0])
    """



from pathlib import Path

import pytest
from qtpy import QtWidgets

from fastmolwidget.cif.cif_file_io import CifReader
from fastmolwidget.molecule2D import calc_volume, RenderItem, MoleculeWidget
from fastmolwidget.sdm import Atomtuple

app = QtWidgets.QApplication.instance()
if not app:
    app = QtWidgets.QApplication([])
data = Path('tests/test-data')


def test_calc_volume():
    # Test with orthogonal cell (e.g., cubic 10, 10, 10, 90, 90, 90)
    vol = calc_volume(10.0, 10.0, 10.0, 90.0, 90.0, 90.0)
    assert vol == pytest.approx(1000.0, rel=1e-5)

    # Test with monoclinic cell
    vol = calc_volume(10.0, 10.0, 10.0, 90.0, 120.0, 90.0)
    assert vol == pytest.approx(866.0254, rel=1e-5)


def test_render_item():
    item = RenderItem(is_bond=True, z_order=1.5)
    assert item.is_bond is True
    assert item.z_order == 1.5
    assert item.atom1 is None


def test_molecule_widget_creation():
    widget = MoleculeWidget()
    assert widget.atoms_size == 12
    assert widget.fontsize == 13
    assert widget.bond_width == 3
    assert widget.labels is True
    assert widget._show_adps is True
    assert widget.bond_drawer == widget._draw_bond_rounded


def test_adp_intersection_line_width_scales_with_zoom():
    widget = MoleculeWidget()
    widget._factor = 0.2
    thin = widget._adp_intersection_line_width()
    widget._factor = 1.0
    thick = widget._adp_intersection_line_width()
    assert thick > thin


def test_molecule_widget_with_cif():
    cif = CifReader(data / '1979688_small.cif')
    adp_dict = {dp.label: (dp.U11, dp.U22, dp.U33, dp.U23, dp.U13, dp.U12) for dp in cif.displacement_parameters()}

    widget = MoleculeWidget()
    widget.resize(800, 600)
    widget.open_molecule(list(cif.atoms_orth), cif.cell[:6], adp_dict)
    widget.show()

    assert len(widget.atoms) == 94

    clicked_atom = widget.atoms[7]
    clicked_atom.screenx = 80
    clicked_atom.screeny = 222
    assert widget.is_point_inside_atom(clicked_atom, 80.0, 222) == True

    # Ensure grabbing the widget content as pixmap (invoking paintEvent) does not crash
    pixmap = widget.grab()
    assert not pixmap.isNull()

    # Test setting parameters and re-drawing
    widget.labels = False
    widget.show_adps = False
    widget.atoms_size = 15
    widget.bond_width = 4
    widget.repaint()

    # Test grabbing again to ensure settings do not crash rendering
    pixmap_updated = widget.grab()
    assert not pixmap_updated.isNull()

    # Test interaction (zooming, reset)
    widget.reset_view()
    widget.zoom = 1.2

    pixmap_transformed = widget.grab()
    assert not pixmap_transformed.isNull()


def test_molecule_widget_toggles():
    widget = MoleculeWidget()

    # Test setting label visibility
    widget.set_labels_visible(False)
    assert widget.labels is False
    widget.show_labels(True)
    assert widget.labels is True

    # Test hydrogen visibility
    widget.show_hydrogens(False)
    assert widget.show_hydrogens_flag is False

    # Test ADP visibility
    widget.show_adps(False)
    assert widget._show_adps is False

    # Test bond type toggle
    widget.show_round_bonds(False)
    assert widget.bond_drawer == widget._draw_bond
    widget.show_round_bonds(True)
    assert widget.bond_drawer == widget._draw_bond_rounded

    # Test label font setting
    widget.setLabelFont(20)
    assert widget.fontsize == 20
    widget.setLabelFont(-5)
    assert widget.fontsize == 1

    # Test set background color
    from qtpy.QtGui import QColor, QPalette
    from qtpy import QtCore
    widget.set_background_color(QColor(QtCore.Qt.GlobalColor.black))
    assert widget.palette().color(QPalette.ColorRole.Window).name() == QColor(QtCore.Qt.GlobalColor.black).name()


def test_molecule_widget_clear():
    widget = MoleculeWidget()

    # create dummy atoms
    dummy_atom = Atomtuple('C1', 'C', 0.0, 0.0, 0.0, 0)
    widget.open_molecule([dummy_atom])
    assert len(widget.atoms) == 1

    assert widget.is_point_inside_atom(widget.atoms[0], 0, 0) == True
    assert widget.is_point_inside_atom(widget.atoms[0], 100, 100) == False

    widget.clear()
    assert len(widget.atoms) == 0


def test_molecule_widget_rotation_matrices():
    widget = MoleculeWidget()
    widget.x_angle = 3.14159 / 2  # 90 degrees approx
    widget.y_angle = 3.14159 / 2

    rx = widget.rotate_x()
    ry = widget.rotate_y()

    import numpy as np
    assert rx.shape == (3, 3)
    assert ry.shape == (3, 3)
    # just checking that they run and return a matrix
    assert isinstance(rx, np.ndarray)


def test_mouse_events_record_position():
    widget = MoleculeWidget()
    widget.resize(200, 200)
    widget.show()

    from qtpy.QtCore import QPointF, QPoint
    from qtpy import QtCore
    from qtpy.QtTest import QTest

    QTest.mousePress(widget, QtCore.Qt.MouseButton.LeftButton,
                     QtCore.Qt.KeyboardModifier.NoModifier, QPoint(10, 20))

    assert widget._lastPos == QPointF(10.0, 20.0)
    assert widget._pressPos == QPointF(10.0, 20.0)



from fastmolwidget.viewer_widget import MoleculeViewerWidget
from fastmolwidget.molecule2D import MoleculeWidget
from fastmolwidget.loader import MoleculeLoader

__all__ = ["MoleculeViewerWidget", "MoleculeWidget", "MoleculeLoader"]


def main() -> None:
    print("Hello from fastmolwidget!")



import shutil
import unittest
from pathlib import Path

from fastmolwidget.cif.cif_file_io import CifReader

data = Path('tests/test-data')


class CifFileCRCTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.cif = CifReader(data / '1979688.cif')

    def test_calc_crc(self):
        self.assertEqual(20714, self.cif.calc_checksum(self.cif['_shelx_hkl_file']))


class CifFileTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.cif = CifReader(data / '1979688.cif')

    def test_calc_crc(self):
        self.assertEqual(3583, self.cif.calc_checksum('hello world'))

    def test_res_crc(self):
        self.assertEqual(17612, self.cif.res_checksum_calcd)

    def test_hkl_crc(self):
        self.assertEqual(20714, self.cif.hkl_checksum_calcd)

    def test_res_crc_without_res(self):
        self.assertEqual(0, CifReader(data / '1000006.cif').res_checksum_calcd)

    def test_get_unknown_value_from_key(self):
        self.assertEqual('', self.cif['_chemical_melting_point'])

    def test_get_known_value_from_key(self):
        self.assertEqual('702.70', self.cif['_chemical_formula_weight'])

    def test_get_spgr(self):
        self.assertEqual('P 21 21 2', self.cif.space_group)

    def test_symmops(self):
        self.assertEqual(['x, y, z', '-x, -y, z', '-x+1/2, y+1/2, -z', 'x+1/2, -y+1/2, -z'], self.cif.symmops)

    def test_symmops_from_spgr(self):
        self.assertEqual(['x,y,z', '-x,-y,z', 'x+1/2,-y+1/2,-z', '-x+1/2,y+1/2,-z'], self.cif.symmops_from_spgr)

    def test_centrosymm(self):
        self.assertEqual(False, self.cif.is_centrosymm)
        c = CifReader(data / 'p21c.cif')
        self.assertEqual(True, c.is_centrosymm)

    def test_ishydrogen(self):
        self.assertEqual(True, self.cif.ishydrogen('H18a'))
        self.assertEqual(True, self.cif.ishydrogen('H18A'))
        self.assertEqual(False, self.cif.ishydrogen('C2'))
        self.assertEqual(False, self.cif.ishydrogen('c2'))

    def test_cell(self):
        expected = [round(x, 8) for x in (19.678, 37.02290000000001, 4.772, 90.0, 90.0, 90.0, 3476.576780226401)]
        actual = [round(y, 8) for y in self.cif.cell]
        self.assertEqual(expected, actual)

    def test_natoms(self):
        self.assertEqual(94, self.cif.natoms())
        self.assertEqual(52, self.cif.natoms(without_h=True))

    def test_checksum_tests(self):
        self.assertEqual(True, self.cif.test_hkl_checksum())
        self.assertEqual(True, self.cif.test_res_checksum())

    def test_checksum_test_without_checksum(self):
        self.assertEqual(True, CifReader(data / '1000006.cif').test_res_checksum())
        self.assertEqual(True, CifReader(data / '1000006.cif').test_hkl_checksum())

    def test_distance_from_string(self):
        self.assertEqual('1.527(3)', self.cif.bond_dist('C1-C2'))

    def test_distance_from_false_string(self):
        self.assertEqual(None, self.cif.bond_dist('C1-C999'))

    def test_angle_from_string(self):
        self.assertEqual('109.9', self.cif.angle('C1-C2-H2'))

    def test_angle_from_false_string(self):
        self.assertEqual(None, self.cif.angle('C1-C2-Hxx'))

    def test_torsion_angle_from_string(self):
        self.assertEqual('173.5(2)', self.cif.torsion('C1-C2-C3-C4'))

    def test_torsion_angle_from_false_string(self):
        self.assertEqual(None, self.cif.torsion('C1-C2-C3-C999'))

    def test_add_loop(self):
        self.assertEqual(9, self.cif.n_loops)
        self.cif.add_loop_to_cif(loop_tags=['_foo', '_bar'], row_values=['fooval', 'barval'])
        self.assertEqual(10, self.cif.n_loops)
        self.assertEqual(['fooval', 'barval'], self.cif.get_loop('_foo').values)

    def test_add_complete_loop(self):
        loop_data = {'_foo_bar': ['1', '2', '3', '4'], '_baz_hhk': ['omega', 'phi', 'phi', 'chi']}
        self.cif.add_loop_from_columns(list(loop_data.keys()), list(loop_data.values()))
        self.assertEqual(['1', 'omega', '2', 'phi', '3', 'phi', '4', 'chi'],
                         self.cif.get_loop('_foo_bar').values)

    def test_publ_flag_not_set(self):
        self.assertEqual("bond(label1='C1', label2='O1', dist='1.438(3)', symm='.')", str(next(iter(self.cif.bonds()))))
        self.cif.block.find_loop('_geom_bond_publ_flag')[0] = 'no'
        # This is not C1-O1, but C1-C14, because the first bond is with publ_flag 'no'.
        self.assertEqual("bond(label1='C1', label2='C14', dist='1.519(3)', symm='.')",
                         str(next(iter(self.cif.bonds()))))
        self.cif.block.find_loop('_geom_bond_publ_flag').erase()
        self.assertEqual([], list(self.cif.block.find_loop('_geom_bond_publ_flag')))
        self.assertEqual("bond(label1='C1', label2='O1', dist='1.438(3)', symm='.')", str(next(iter(self.cif.bonds()))))


class TestQuotationMark(unittest.TestCase):
    def setUp(self) -> None:
        self.cif = CifReader(data / '1979688.cif')
        shutil.copyfile(self.cif.fileobj, data / 'test.cif')
        self.cif2 = CifReader(data / 'test.cif')

    def tearDown(self) -> None:
        self.cif2.fileobj.unlink(missing_ok=True)

    def test_insert_quotation_mark(self):
        self.cif['_diffrn_detector'] = "Jesus' live"
        self.assertEqual("Jesus' live", self.cif['_diffrn_detector'])
        self.cif.save(self.cif2.fileobj)
        self.cif2 = CifReader(self.cif2.fileobj)
        self.assertEqual("Jesus' live", self.cif['_diffrn_detector'])


class TestMultiCif(unittest.TestCase):
    def setUp(self) -> None:
        self.cif = CifReader(data / 'multi.cif')

    def test_ismulti(self):
        self.assertEqual(True, self.cif.is_multi_cif)

    def test_isempty(self):
        self.assertEqual(False, self.cif.is_empty())


if __name__ == '__main__':
    unittest.main()



"""
A self-contained QWidget that embeds a :class:`~fastmolwidget.molecule2D.MoleculeWidget`
together with its control bar.

Usage::

    viewer = MoleculeViewerWidget()
    viewer.load_file("structure.cif")
    viewer.show()

"""

from __future__ import annotations

from pathlib import Path

from qtpy import QtWidgets

from fastmolwidget.loader import MoleculeLoader
from fastmolwidget.molecule2D import MoleculeWidget


class MoleculeViewerWidget(QtWidgets.QWidget):
    """A ready-to-use viewer widget that combines a :class:`MoleculeWidget`
    with a control bar (ADP toggle, label toggle, bond style, hydrogen
    visibility, and bond-width adjustment).

    The interface is intentionally minimal: call :meth:`load_file` to display a
    structure.

    :param parent: Optional parent widget.
    """

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        # ── molecule renderer ────────────────────────────────────────────────
        self._render_widget = MoleculeWidget()
        self._loader = MoleculeLoader(self._render_widget)

        # ── control bar ──────────────────────────────────────────────────────
        self._adp_checkbox = QtWidgets.QCheckBox("Show ADP")
        self._label_checkbox = QtWidgets.QCheckBox("Show Labels")
        self._bond_type_checkbox = QtWidgets.QCheckBox("Round Bonds")
        self._hydrogens_checkbox = QtWidgets.QCheckBox("Show Hydrogens")
        self._grow_checkbox = QtWidgets.QCheckBox("Grow")

        self._bw_label = QtWidgets.QLabel("Bond Width:")
        self._bond_width_spinbox = QtWidgets.QSpinBox()
        self._bond_width_spinbox.setRange(1, 15)
        self._bond_width_spinbox.setValue(3)

        # default state
        self._adp_checkbox.setChecked(True)
        self._bond_type_checkbox.setChecked(True)
        self._hydrogens_checkbox.setChecked(True)

        # wire controls to renderer
        self._adp_checkbox.toggled.connect(self._render_widget.show_adps)
        self._label_checkbox.toggled.connect(self._render_widget.show_labels)
        self._bond_type_checkbox.toggled.connect(self._render_widget.show_round_bonds)
        self._hydrogens_checkbox.toggled.connect(self._render_widget.show_hydrogens)
        self._bond_width_spinbox.valueChanged.connect(self._render_widget.set_bond_width)
        self._grow_checkbox.toggled.connect(self._loader.set_grow)

        # apply initial defaults to the renderer
        self._render_widget.set_bond_width(3)
        self._render_widget.show_labels(False)
        self._render_widget.show_round_bonds(True)

        # ── layout ───────────────────────────────────────────────────────────
        control_bar = QtWidgets.QHBoxLayout()
        control_bar.addWidget(self._grow_checkbox)
        control_bar.addWidget(self._adp_checkbox)
        control_bar.addWidget(self._label_checkbox)
        control_bar.addWidget(self._bond_type_checkbox)
        control_bar.addWidget(self._hydrogens_checkbox)
        control_bar.addWidget(self._bw_label)
        control_bar.addWidget(self._bond_width_spinbox)
        control_bar.addStretch()

        vl = QtWidgets.QVBoxLayout(self)
        vl.addWidget(self._render_widget)
        vl.addLayout(control_bar)

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    @property
    def render_widget(self) -> MoleculeWidget:
        """The underlying :class:`MoleculeWidget` (read-only)."""
        return self._render_widget

    def load_file(self, filename: str | Path) -> None:
        """Load a structure file and display it.

        The file format is determined from the extension (``.cif``, ``.res``,
        ``.ins``, ``.xyz``).

        :param filename: Path to the structure file.
        :raises ValueError: If the file format is not supported.
        :raises FileNotFoundError: If the file does not exist.
        """
        self._loader.load_file(filename)

if __name__ == '__main__':

    app = QtWidgets.QApplication.instance()
    if not app:
        app = QtWidgets.QApplication([])

    w = MoleculeViewerWidget()
    #w.load_file('../../tests/test-data/p31c.cif')
    w.load_file('../../tests/test-data/p31c-finalcif.res')
    w.show()
    app.exec()


"""Tests for :class:`~fastmolwidget.viewer_widget.MoleculeViewerWidget`."""

from pathlib import Path

import pytest
from qtpy import QtWidgets

from fastmolwidget.viewer_widget import MoleculeViewerWidget

app = QtWidgets.QApplication.instance()
if not app:
    app = QtWidgets.QApplication([])

data = Path('tests/test-data')


# ------------------------------------------------------------------
# Construction
# ------------------------------------------------------------------

def test_construction_no_args():
    w = MoleculeViewerWidget()
    assert w is not None
    assert w.render_widget is not None


# ------------------------------------------------------------------
# load_file
# ------------------------------------------------------------------

def test_load_cif():
    w = MoleculeViewerWidget()
    w.load_file(data / '1979688_small.cif')
    assert len(w.render_widget.atoms) == 94


def test_load_xyz():
    w = MoleculeViewerWidget()
    w.load_file(data / 'test_molecule.xyz')
    assert len(w.render_widget.atoms) == 5


def test_load_shelx():
    w = MoleculeViewerWidget()
    w.load_file(data / 'test_molecule.res')
    assert len(w.render_widget.atoms) == 5


def test_load_unsupported_format():
    w = MoleculeViewerWidget()
    with pytest.raises(ValueError, match='Unsupported file format'):
        w.load_file(data / 'fake.pdb')


def test_load_missing_file():
    w = MoleculeViewerWidget()
    with pytest.raises(FileNotFoundError):
        w.load_file('nonexistent.cif')


# ------------------------------------------------------------------
# Rendering sanity check
# ------------------------------------------------------------------

def test_viewer_renders():
    w = MoleculeViewerWidget()
    w.load_file(data / '1979688_small.cif')
    # app.processEvents()
    assert w.render_widget.width() > 0
    assert w.render_widget.height() > 0
    pixmap = w.grab()
    assert not pixmap.isNull()



from __future__ import annotations

from fastmolwidget.atoms import get_radius_from_element, element2color
from fastmolwidget.sdm import Atomtuple

"""
A versatile 2D/3D molecule drawing widget for PyQt/PySide.
Renders molecules as ORTEP-style thermal ellipsoid plots (when anisotropic
displacement parameters are provided) or as simple ball-and-stick diagrams.
The widget supports interactive mouse rotation, zooming, and panning.

Feed it with a list of :class:`~finalcif.displaymol.sdm.Atomtuple` objects
whose fields are:

- **label**:       Atom name, e.g. ``'C3'``.
- **type**:        Element symbol as string, e.g. ``'C'``.
- **x, y, z**:    Atom position in Cartesian coordinates (Ångströms).
- **part**:        Disorder part in SHELX notation (0 = all parts, ..., -1, -2, 1, 2, ...).
- **symm_matrix**: Optional 3×3 fractional rotation matrix used to rotate
  the ADP tensor for symmetry-generated atoms.

Mouse controls (inside :class:`MoleculeWidget`):

- **Left drag**:   Rotate the molecule.
- **Right drag**:  Zoom in / out.
- **Middle drag**: Pan the view.
- **Scroll wheel**: Increase / decrease label font size.
- **Left click**:  Select a single atom or bond (clears previous selection).
- **Ctrl + Left click**: Toggle selection of multiple atoms and bonds.
"""

from dataclasses import dataclass
from math import sqrt, cos, sin, dist, radians, atan2, degrees, pi
from pathlib import Path

import numpy as np
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt, QRectF
from qtpy.QtGui import QPainter, QPen, QBrush, QColor, QMouseEvent, QPalette, QImage, QResizeEvent, QWheelEvent, \
    QRadialGradient, QLinearGradient, QTransform


def calc_volume(a: float, b: float, c: float, alpha: float, beta: float, gamma: float) -> float:
    ca, cb, cg = cos(radians(alpha)), cos(radians(beta)), cos(radians(gamma))
    return a * b * c * sqrt(1 + 2 * ca * cb * cg - ca ** 2 - cb ** 2 - cg ** 2)


@dataclass(slots=True)
class RenderItem:
    """A single renderable element (atom or bond) in the z-ordered draw list.

    :param is_bond: ``True`` for a bond, ``False`` for an atom.
    :param z_order: Depth value used to sort back-to-front for the painter's algorithm.
    :param atom1: The first (or only) atom involved.
    :param atom2: The second atom of a bond, or ``None`` for an atom item.
    """

    is_bond: bool
    z_order: float = 0.0
    atom1: Atom = None
    atom2: Atom | None = None


class MoleculeWidget(QtWidgets.QWidget):
    """Interactive Qt widget that renders a molecule as a 2D projection.

    Supports ORTEP-style anisotropic displacement parameter (ADP) ellipsoids
    at 50 % probability level, isotropic spheres, and simple ball-and-stick
    representations.  The molecule can be rotated (left-drag), zoomed
    (right-drag), and panned (middle-drag) with the mouse.

    Typical usage::

        widget = MoleculeWidget(parent)
        widget.open_molecule(atoms=atom_list, cell=cell_params, adps=adp_dict)

    :param parent: Optional parent widget.
    """

    atomClicked = QtCore.Signal(str)
    bondClicked = QtCore.Signal(str, str)

    def __init__(self, parent: QtGui.QWidget = None):
        super().__init__(parent)
        self._astar = None
        self._bstar = None
        self._cstar = None
        self._amatrix = None
        self._adp_map = None
        self._cell = None
        self._factor = 1.0
        self.atoms_size = 12
        self.fontsize = 13
        self.bond_width = 3
        self.labels = True
        self._show_adps = True
        self.bond_drawer = self._draw_bond_rounded

        self.show_hydrogens_flag = True

        # Track selected atoms and bonds as sets for multi-selection
        self.selected_atoms: set[str] = set()
        self.selected_bonds: set[tuple[str, str]] = set()

        # scaling factor for ADP ellipsoids in screen coordinates
        # 1.5382 is the standard ORTEP scaling factor for 50% probability:
        self.adp_scale = 1.5382
        self.molecule_center = np.array([0, 0, 0])
        self.molecule_radius = 10
        self._lastPos = self.pos()
        self._pressPos = self.pos()
        self._painter: None | QPainter = None
        self.x_angle = 0
        self.y_angle = 0

        self.scale = 150.0
        self.cx_global = 0.0
        self.cy_global = 0.0

        # Cumulative rotation matrix to preserve orientation during grow
        self.cumulative_R = np.eye(3, dtype=np.float32)

        # Color caches
        self.bond_color = QColor('#555555')
        self.fallback_pen_color = QColor(QtCore.Qt.GlobalColor.black)
        self.adp_pen_color = QColor(0, 0, 0, 255)

        # 3D Cylinder gradient colors for bonds
        self.bond_grad_dark = QColor(60, 60, 60)
        self.bond_grad_light = QColor(140, 140, 140)
        self.bond_grad_shadow = QColor(10, 10, 10)

        bg = QLinearGradient(0, 1, 0, -1)
        bg.setColorAt(0.0, self.bond_grad_dark)
        bg.setColorAt(0.2, self.bond_grad_light)
        bg.setColorAt(1.0, self.bond_grad_shadow)
        self.bond_brush = QBrush(bg)

        pal = QPalette()
        pal.setColor(QtGui.QPalette.ColorRole.Window, QtCore.Qt.GlobalColor.white)
        self.setAutoFillBackground(True)
        self.setPalette(pal)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        self.atoms: list[Atom] = []
        self.connections = ()
        self.objects: list[RenderItem] = []

        self.screen_center = [self.width() / 2, self.height() / 2]
        self.zoom = 1.1

        # Numpy arrays for fast vectorized rotation
        self._coords_array = np.empty((0, 3))
        self._ucart_array = np.empty((0, 3, 3))
        self._has_adp = np.empty(0, dtype=bool)
        self._eigenvalues_array = np.empty((0, 3))
        self._eigenvectors_array = np.empty((0, 3, 3))
        self._u_inv_array = np.empty((0, 3, 3))

        self.mouse_pressed = False
        self.labels = True
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.atomClicked.connect(lambda x: print(f"Atom Selected: {x}"))
        self.bondClicked.connect(lambda x, y: print(f"Bond Selected: {x}-{y}"))

    def set_background_color(self, color: QColor):
        """Set the background color of the widget."""
        self.bg_color = color
        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Window, color)
        self.setPalette(palette)

    def sizeHint(self) -> QtCore.QSize:
        """Preferred starting size so the render area stays visible in layouts."""
        return QtCore.QSize(640, 480)

    def minimumSizeHint(self) -> QtCore.QSize:
        """Reasonable minimum size for molecule rendering."""
        return QtCore.QSize(320, 220)

    def _adp_intersection_line_width(self) -> float:
        """Return a zoom-aware stroke width for ADP intersection lines."""
        return max(1.0, min(6.0, self._factor * 3.0))

    def set_bond_width(self, width: int):
        """Set the width of the bonds."""
        self.bond_width = width
        self.update()

    def set_labels_visible(self, visible: bool):
        """Toggle visibility of atom labels."""
        self.labels = visible
        self.update()

    def show_hydrogens(self, value: bool):
        """Toggle display of hydrogen atoms and their bonds."""
        self.show_hydrogens_flag = value
        self.update()

    def reset_view(self):
        """Reset zoom and rotation to defaults."""
        self.zoom = 1.0
        self.x_angle = 180.0
        self.y_angle = 180.0
        self.z_angle = 0.0
        self.x_shift_screen = 0
        self.y_shift_screen = 0
        self.cumulative_R = np.eye(3, dtype=np.float32)
        self.update()

    def setLabelFont(self, font_size: int):
        """Set the pixel size used for atom labels and schedule a repaint."""
        if font_size < 0:
            font_size = 1
        self.fontsize = font_size
        self.update()

    def clear(self) -> None:
        """Remove all atoms and bonds from the widget."""
        self.open_molecule(atoms=[])

    def show_labels(self, value: bool):
        """Toggle the display of non-hydrogen atom labels."""
        self.labels = value
        self.update()

    def show_adps(self, value: bool):
        """Toggle the display of ADP ellipsoids / isotropic spheres."""
        self._show_adps = value
        self.update()

    def show_round_bonds(self, bond_type: bool = True):
        """Switch between flat and 3D-shaded (rounded) bond rendering."""
        if not bond_type:
            self.bond_drawer = self._draw_bond
        else:
            self.bond_drawer = self._draw_bond_rounded
        self.update()

    def open_molecule(self,
                      atoms: list[Atomtuple],
                      cell: tuple[float, float, float, float, float, float] | None = None,
                      adps: dict[str, tuple[float, float, float, float, float, float]] | None = None,
                      keep_view: bool = False) -> None:
        """
        Loads a new molecule and completely resets the view (zoom, pan, rotation).
        """
        self._load_molecule(atoms, cell, adps, keep_view=keep_view)

    def grow_molecule(self,
                      atoms: list[Atomtuple],
                      cell: tuple[float, float, float, float, float, float] | None = None,
                      adps: dict[str, tuple[float, float, float, float, float, float]] | None = None) -> None:
        """Updates the molecule while preserving the current view (zoom, pan, rotation)."""
        self._load_molecule(atoms, cell, adps, keep_view=True)

    def _load_molecule(self, atoms: list[Atomtuple],
                       cell: tuple[float, float, float, float, float, float] | None = None,
                       adps: dict[str, tuple[float, float, float, float, float, float]] | None = None,
                       keep_view: bool = False) -> None:

        self._cell = cell
        self._adp_map = adps if adps is not None else {}

        if self._cell is not None and self._show_adps:
            self.calc_amatrix()

        self.atoms.clear()
        self.make_adps(atoms)
        self.connections = self.get_conntable_from_atoms()

        if not keep_view:
            self.get_center_and_radius()
            self.cumulative_R = np.eye(3, dtype=np.float32)
            self.selected_atoms.clear()
            self.selected_bonds.clear()

        self.objects.clear()
        for n1, n2 in self.connections:
            at1 = self.atoms[n1]
            at2 = self.atoms[n2]

            self.objects.append(RenderItem(is_bond=True, atom1=at1, atom2=at2))

        for atom in self.atoms:
            if atom.type_ in ('H', 'D'):
                atom.u_iso = 0.01
            self.objects.append(RenderItem(is_bond=False, atom1=atom))

        # Build numpy arrays for fully vectorized rotation
        self._coords_array = np.array([at.coordinate for at in self.atoms])
        self._ucart_array = np.zeros((len(self.atoms), 3, 3))
        self._has_adp = np.zeros(len(self.atoms), dtype=bool)
        self._eigenvalues_array = np.zeros((len(self.atoms), 3))
        self._eigenvectors_array = np.zeros((len(self.atoms), 3, 3))
        self._u_inv_array = np.zeros((len(self.atoms), 3, 3))

        # Apply accumulated rotation to the new coordinates
        if keep_view and not np.allclose(self.cumulative_R, np.eye(3)):
            self._coords_array = np.dot(self._coords_array - self.molecule_center,
                                        self.cumulative_R.T) + self.molecule_center

        for i, at in enumerate(self.atoms):
            if keep_view and not np.allclose(self.cumulative_R, np.eye(3)):
                at.coordinate = self._coords_array[i]

            at.z = at.coordinate[2]
            if at.u_cart is not None:
                if keep_view and not np.allclose(self.cumulative_R, np.eye(3)):
                    # Rotate the ADP tensor before computing eigenvectors/values
                    at.u_cart = np.matmul(self.cumulative_R, np.matmul(at.u_cart, self.cumulative_R.T))

                try:
                    evals, evecs = np.linalg.eigh(at.u_cart)
                    if np.any(evals <= 0):
                        at.adp_valid = False
                    else:
                        at.adp_valid = True
                    u_invers = np.linalg.inv(at.u_cart)

                    self._ucart_array[i] = at.u_cart
                    self._eigenvalues_array[i] = evals
                    self._eigenvectors_array[i] = evecs
                    self._u_inv_array[i] = u_invers
                    self._has_adp[i] = True

                    at.u_eigvals = evals
                    at.u_eigvecs = evecs
                    at.u_inv = u_invers
                except np.linalg.LinAlgError:
                    at.adp_valid = False
                    at.u_cart = None
                    at.u_iso = None

        if not keep_view:
            self._factor = min(self.width(), self.height()) / 2 / self.molecule_radius * self.zoom / 100
            self.atoms_size = self._factor * 70

        self.update()

    def calc_amatrix(self):
        """Compute the orthogonalisation matrix and reciprocal-lattice lengths."""
        a, b, c, alpha, beta, gamma = self._cell
        V = calc_volume(a, b, c, alpha, beta, gamma)
        # reciprocal lattice vectors
        self._astar = (b * c * sin(radians(alpha))) / V
        self._bstar = (c * a * sin(radians(beta))) / V
        self._cstar = (a * b * sin(radians(gamma))) / V
        # orthogonalization matrix (fractional -> cartesian)
        self._amatrix = np.array([
            [a, b * cos(radians(gamma)), c * cos(radians(beta))],
            [0, b * sin(radians(gamma)),
             c * (cos(radians(alpha)) - cos(radians(beta)) * cos(radians(gamma))) / sin(radians(gamma))],
            [0, 0, V / (a * b * sin(radians(gamma)))]
        ], dtype=float)

    def make_adps(self, atoms: list[Atomtuple]) -> None:
        """Convert Atomtuples to internal :class:`Atom` objects with Cartesian ADPs."""
        self.atoms.clear()
        name_counts = {}

        for at in atoms:
            base_name = at.label
            count = name_counts.get(base_name, 0)

            if count == 0:
                internal_name = base_name
            else:
                internal_name = f"{base_name}>>{count}"

            name_counts[base_name] = count + 1

            a = Atom(at.x, at.y, at.z, internal_name, at.type, at.part)
            if self._adp_map and self._cell and base_name in self._adp_map:
                try:
                    uvals = self._adp_map[base_name]
                    symm_matrix = getattr(at, 'symm_matrix', None)
                    if symm_matrix is not None:
                        symm_matrix = np.array(symm_matrix, dtype=float)
                    a.u_cart = self._uij_to_cart(uvals, symm_matrix)
                    a.u_iso = np.trace(a.u_cart) / 3.0
                except Exception:
                    a.u_cart = None
                    a.u_iso = None
            self.atoms.append(a)

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        """Repaint the widget by re-rendering the molecule scene."""
        if self.atoms:
            self._painter = QPainter(self)
            self._painter.setRenderHint(QPainter.RenderHint.Antialiasing)
            font = self._painter.font()
            font.setPixelSize(self.fontsize)
            self._painter.setFont(font)
            try:
                self.draw()
            except (ValueError, IndexError) as e:
                print(f'Draw structure crashed: {e}')
                self._painter.end()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Record the initial cursor position for subsequent drag operations."""
        self._lastPos = event.position()
        self._pressPos = event.position()

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        """Detect clicks on atoms or bonds and emit corresponding signals."""
        if event.button() == QtCore.Qt.MouseButton.LeftButton:
            dx = event.position().x() - self._pressPos.x()
            dy = event.position().y() - self._pressPos.y()
            if abs(dx) < 5 and abs(dy) < 5:
                x = event.position().x()
                y = event.position().y()

                clicked_atom = None
                clicked_bond = None
                front_z = float('inf')

                # Check all objects to find the single one that is most in front
                for item in self.objects:
                    # Skip hidden hydrogens
                    if not self.show_hydrogens_flag:
                        if item.atom1.type_ in ('H', 'D') or (item.is_bond and item.atom2.type_ in ('H', 'D')):
                            continue

                    if item.is_bond:
                        if self.is_point_near_bond(item.atom1, item.atom2, x, y):
                            if item.z_order < front_z:
                                front_z = item.z_order
                                clicked_bond = tuple(sorted((item.atom1.name, item.atom2.name)))
                                clicked_atom = None
                    else:
                        if self.is_point_inside_atom(item.atom1, x, y):
                            if item.z_order < front_z:
                                front_z = item.z_order
                                clicked_atom = item.atom1
                                clicked_bond = None

                # Update selection state
                modifiers = event.modifiers()
                ctrl_pressed = bool(modifiers & Qt.KeyboardModifier.ControlModifier)
                selection_changed = False

                if clicked_atom:
                    if ctrl_pressed:
                        if clicked_atom.name in self.selected_atoms:
                            self.selected_atoms.remove(clicked_atom.name)
                        else:
                            self.selected_atoms.add(clicked_atom.name)
                    else:
                        self.selected_atoms = {clicked_atom.name}
                        self.selected_bonds.clear()

                    selection_changed = True
                    self.atomClicked.emit(clicked_atom.name)

                elif clicked_bond:
                    if ctrl_pressed:
                        if clicked_bond in self.selected_bonds:
                            self.selected_bonds.remove(clicked_bond)
                        else:
                            self.selected_bonds.add(clicked_bond)
                    else:
                        self.selected_bonds = {clicked_bond}
                        self.selected_atoms.clear()

                    selection_changed = True
                    self.bondClicked.emit(clicked_bond[0], clicked_bond[1])

                else:
                    if not ctrl_pressed and (self.selected_atoms or self.selected_bonds):
                        self.selected_atoms.clear()
                        self.selected_bonds.clear()
                        selection_changed = True

                if selection_changed:
                    self.update()

        super().mouseReleaseEvent(event)

    def is_point_inside_atom(self, atom: Atom, px: float, py: float) -> bool:
        """Check if a screen coordinate is inside the atom's drawn 2D projection."""
        cx = atom.screenx
        cy = atom.screeny
        dx = px - cx
        dy = py - cy

        if self._show_adps and atom.u_cart is not None:
            a = atom.u_cart[0, 0]
            b = atom.u_cart[0, 1]
            c = atom.u_cart[1, 1]

            T = a + c
            D = a * c - b * b
            diff = T * T * 0.25 - D

            if diff >= 0:
                sq = sqrt(diff)
                eig1 = T * 0.5 - sq
                eig2 = T * 0.5 + sq

                if eig1 > 0 and eig2 > 0:
                    r1 = sqrt(eig1) * self.scale * self.adp_scale
                    r2 = sqrt(eig2) * self.scale * self.adp_scale

                    if abs(b) > 1e-8:
                        angle = degrees(atan2(eig1 - a, b))
                    else:
                        angle = 0.0 if a < c else 90.0

                    rad = radians(angle)
                    cos_a = cos(rad)
                    sin_a = sin(rad)

                    # Transform to ellipse local coordinates
                    local_x = dx * cos_a + dy * sin_a
                    local_y = -dx * sin_a + dy * cos_a

                    if (local_x ** 2 / r1 ** 2) + (local_y ** 2 / r2 ** 2) <= 1.0:
                        return True
                    return False

        # Fallback for isotropic spheres or fixed-size circles
        circle_size = self.atoms_size
        if self._show_adps and atom.u_iso is not None:
            r = sqrt(atom.u_iso) * self.scale * self.adp_scale
            circle_size = r * 2

        radius = circle_size / 2
        return dx ** 2 + dy ** 2 <= radius ** 2

    def is_point_near_bond(self, at1: Atom, at2: Atom, px: float, py: float) -> bool:
        """Check if a screen coordinate is sufficiently close to a drawn bond segment."""
        line_data = self._get_bond_line(at1, at2)
        if not line_data:
            return False

        x1, y1, x2, y2, dynamic_width = line_data

        line_vec = np.array([x2 - x1, y2 - y1])
        p_vec = np.array([px - x1, py - y1])
        line_len_sq = np.dot(line_vec, line_vec)

        if line_len_sq == 0.0:
            return False

        t = max(0, min(1, np.dot(p_vec, line_vec) / line_len_sq))
        proj = np.array([x1, y1]) + t * line_vec
        dist_sq = (px - proj[0]) ** 2 + (py - proj[1]) ** 2

        # Add 4 extra pixels of hit area tolerance around the bond
        tolerance = max(5.0, dynamic_width / 2.0 + 4.0)
        return dist_sq <= tolerance ** 2

    def wheelEvent(self, event: QWheelEvent):
        """Increase or decrease the label font size on scroll."""
        if event.angleDelta().y() > 0:
            self.setLabelFont(self.fontsize + 2)
        elif event.angleDelta().y() < 0:
            self.setLabelFont(self.fontsize - 2)

    def save_image(self, filename: Path, image_scale: float = 1.5) -> None:
        """Render the current molecule view to an image file."""
        image = QImage(self.size() * image_scale, QImage.Format.Format_RGB32)
        image.fill(Qt.GlobalColor.white)
        painter = QPainter(image)
        painter.scale(image_scale, image_scale)
        self.render(painter, QtCore.QPoint(0, 0))
        painter.end()
        image.save(str(filename.resolve()))

    def rotate_x(self) -> np.typing.NDArray[np.float32]:
        """Return a 3×3 rotation matrix around the X axis by :attr:`x_angle` radians."""
        return np.array([
            [1, 0, 0],
            [0, cos(self.x_angle), -sin(self.x_angle)],
            [0, sin(self.x_angle), cos(self.x_angle)],
        ], dtype=np.float32)

    def rotate_y(self) -> np.typing.NDArray[np.float32]:
        """Return a 3×3 rotation matrix around the Y axis by :attr:`y_angle` radians."""
        return np.array([
            [cos(self.y_angle), 0, sin(self.y_angle)],
            [0, 1, 0],
            [-sin(self.y_angle), 0, cos(self.y_angle)],
        ], dtype=np.float32)

    def _uij_to_cart(self, uvals: tuple[float, float, float, float, float, float],
                     symm_matrix: np.ndarray | None = None) -> np.ndarray:
        """Convert fractional *Uij* displacement parameters to a Cartesian tensor."""
        U11, U22, U33, U23, U13, U12 = uvals
        Uij = np.array([[U11, U12, U13],
                        [U12, U22, U23],
                        [U13, U23, U33]], dtype=float)

        # Apply the fractional rotation part of the symmetry operation.
        if symm_matrix is not None:
            Uij = symm_matrix.T @ Uij @ symm_matrix

        N = np.diag([self._astar, self._bstar, self._cstar])
        Ucart = self._amatrix.dot(N).dot(Uij).dot(N.T).dot(self._amatrix.T)
        return Ucart

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        """Dispatch drag events to rotate, zoom, or pan depending on the mouse button."""
        if event.buttons() == QtCore.Qt.MouseButton.LeftButton:
            self.rotate_molecule(event)
        elif event.buttons() == QtCore.Qt.MouseButton.RightButton:
            self.zoom_molecule(event)
        elif event.buttons() == QtCore.Qt.MouseButton.MiddleButton:
            self.pan_molecule(event)
        self._lastPos = event.position()

    def pan_molecule(self, event):
        """Translate the molecule center based on the middle-button drag delta."""
        self.molecule_center[0] += (self._lastPos.x() - event.position().x()) / 50
        self.molecule_center[1] += (self._lastPos.y() - event.position().y()) / 50
        self.update()

    def zoom_molecule(self, event: QMouseEvent):
        """Adjust the zoom / scale factor based on the right-button drag delta."""
        self._factor += (self._lastPos.y() - event.position().y()) / 350
        self._factor = max(0.005, self._factor)
        self.zoom -= (self._lastPos.y() - event.position().y()) / 350
        self.atoms_size = abs(self._factor * 70)
        self.update()

    def rotate_molecule(self, event: QMouseEvent):
        """Rotate the molecule around X and Y axes using the left-button drag delta."""
        self.y_angle = -(event.position().x() - self._lastPos.x()) / 80
        self.x_angle = (event.position().y() - self._lastPos.y()) / 80
        R_y = self.rotate_y()
        R_x = self.rotate_x()
        R = np.dot(R_x, R_y)

        # Track the cumulative rotation matrix
        self.cumulative_R = np.dot(R, self.cumulative_R)

        # Single bulk vector rotation instead of individual loops
        if self.atoms:
            self._coords_array = np.dot(self._coords_array - self.molecule_center, R.T) + self.molecule_center

            if np.any(self._has_adp):
                self._ucart_array = np.matmul(R, np.matmul(self._ucart_array, R.T))
                self._eigenvectors_array = np.matmul(R, self._eigenvectors_array)
                self._u_inv_array = np.matmul(R, np.matmul(self._u_inv_array, R.T))

            for i, at in enumerate(self.atoms):
                at.coordinate = self._coords_array[i]
                at.z = at.coordinate[2]  # cache explicit z property for fast sorting
                if self._has_adp[i]:
                    at.u_cart = self._ucart_array[i]
                    at.u_eigvecs = self._eigenvectors_array[i]
                    at.u_inv = self._u_inv_array[i]

        self.update()

    def get_spherical_radius(self, atom: Atom) -> float:
        """Return an approximate isotropic radius for UI purposes (e.g. label offset)."""
        if self._show_adps and atom.u_iso is not None:
            return sqrt(atom.u_iso)  # * self.adp_scale
        return 0.23

    def get_directional_radius(self, atom: Atom, v: np.ndarray) -> float:
        """Return the distance from the atom centre to its ellipsoid surface along *v*."""
        d = np.linalg.norm(v)
        if d < 1e-8:
            return 0.23

        if not atom.adp_valid:
            return 0.23

        if self._show_adps and atom.u_inv is not None:
            u = v / d
            val = np.dot(u, np.dot(atom.u_inv, u))
            if val > 0:
                return self.adp_scale / sqrt(val)

        # Fallback to sphere
        if self._show_adps and atom.u_iso is not None:
            return sqrt(atom.u_iso) * self.adp_scale

        return 0.23

    def draw(self) -> None:
        """Execute the main rendering pass."""
        self.scale = self._factor * 150
        self.screen_center = [self.width() / 2, self.height() / 2]
        self.cx_global = self.screen_center[0] - self.molecule_center[0] * self.scale
        self.cy_global = self.screen_center[1] - self.molecule_center[1] * self.scale

        hydrogens = ('H', 'D')

        # Precise 2D centers
        for atom in self.atoms:
            c = atom.coordinate
            atom.screenx = c[0] * self.scale + self.cx_global
            atom.screeny = c[1] * self.scale + self.cy_global

        self.calculate_z_order()

        for item in self.objects:
            if not self.show_hydrogens_flag:
                if item.atom1.type_ in hydrogens or (item.is_bond and item.atom2.type_ in hydrogens):
                    continue
            if item.is_bond:
                self.bond_drawer(item.atom1, item.atom2)
            else:
                self.draw_atom(item.atom1)
                if self.labels and item.atom1.type_ not in hydrogens:
                    self.draw_label(item.atom1)
        self._painter.end()

    def calculate_z_order(self):
        """Sort :attr:`objects` back-to-front by depth for the painter's algorithm."""
        for item in self.objects:
            if item.is_bond:
                item.z_order = (item.atom1.z + item.atom2.z) / 2.0
            else:
                item.z_order = item.atom1.z

        self.objects.sort(reverse=True, key=lambda item: item.z_order)

    def get_center_and_radius(self):
        """Compute the bounding sphere of the current atom set."""
        min_ = [999999, 999999, 999999]
        max_ = [-999999, -999999, -999999]
        for at in self.atoms:
            for j in reversed(range(3)):
                v = at.coordinate[j]
                min_[j] = min(min_[j], v)
                max_[j] = max(max_[j], v)
        c = np.array([0, 0, 0], dtype=np.float32)
        for j in reversed(range(3)):
            c[j] = (max_[j] + min_[j]) / 2
        r = 0
        for atom in self.atoms:
            d = dist(atom.coordinate, c) + 1.5
            r = max(r, d)
        self.molecule_center = np.array(c, dtype=np.float32)
        self.molecule_radius = r or 10

    def _get_bond_line(self, at1: Atom, at2: Atom) -> tuple[float, float, float, float, int] | None:
        """Calculate the 2D projected line segment and width for a bond."""
        c1 = at1.coordinate
        c2 = at2.coordinate
        v = c2 - c1
        d = np.linalg.norm(v)

        r1 = self.get_directional_radius(at1, v)
        r2 = self.get_directional_radius(at2, -v)

        if d <= r1 + r2:
            return None

        v_norm = v / d
        p1 = c1 + v_norm * r1
        p2 = c2 - v_norm * r2

        x1 = p1[0] * self.scale + self.cx_global
        y1 = p1[1] * self.scale + self.cy_global
        x2 = p2[0] * self.scale + self.cx_global
        y2 = p2[1] * self.scale + self.cy_global

        dynamic_width = max(1, int(self.bond_width * self._factor * 5))
        return x1, y1, x2, y2, dynamic_width

    def _draw_bond_selection(self, x1: float, y1: float, x2: float, y2: float, dynamic_width: int):
        """Draws a crisp cyan outline behind a selected bond."""
        sel_width = dynamic_width + max(4, int(12 * self._factor))
        pen = QPen(QColor(0, 190, 255), sel_width, Qt.PenStyle.SolidLine)
        pen.setCapStyle(Qt.PenCapStyle.RoundCap)
        self._painter.setPen(pen)
        self._painter.drawLine(int(x1), int(y1), int(x2), int(y2))

    def _draw_bond_rounded(self, at1: Atom, at2: Atom):
        """Draw a 3D-shaded (cylinder-like) bond between two atoms."""
        line_data = self._get_bond_line(at1, at2)
        if not line_data:
            return

        x1, y1, x2, y2, dynamic_width = line_data

        dx = x2 - x1
        dy = y2 - y1
        length = sqrt(dx * dx + dy * dy)
        if length < 0.0001:
            return

        # Check Selection
        bond_key = tuple(sorted((at1.name, at2.name)))
        if bond_key in self.selected_bonds:
            self._draw_bond_selection(x1, y1, x2, y2, dynamic_width)

        # 2D Normal vector
        nx = -dy / length
        ny = dx / length

        # Align normal to point towards the global light source (Top-Left is approx -1, -1)
        Lx, Ly = -1.0, -1.0
        if (nx * Lx + ny * Ly) < 0:
            nx = -nx
            ny = -ny

        t = QTransform(ny * dynamic_width / 2.0, -nx * dynamic_width / 2.0, nx * dynamic_width / 2.0,
                       ny * dynamic_width / 2.0,
                       x1, y1)
        self.bond_brush.setTransform(t)

        pen = QPen(self.bond_brush, dynamic_width, Qt.PenStyle.SolidLine)
        pen.setCapStyle(Qt.PenCapStyle.RoundCap)  # Creates perfect elliptical intersection blend
        self._painter.setPen(pen)
        self._painter.drawLine(int(x1), int(y1), int(x2), int(y2))

    def _draw_bond(self, at1: Atom, at2: Atom) -> None:
        """Draw a flat single-colour bond between two atoms."""
        line_data = self._get_bond_line(at1, at2)
        if not line_data:
            return

        x1, y1, x2, y2, dynamic_width = line_data

        # Check Selection
        bond_key = tuple(sorted((at1.name, at2.name)))
        if bond_key in self.selected_bonds:
            self._draw_bond_selection(x1, y1, x2, y2, dynamic_width)

        pen = QPen(self.bond_color, dynamic_width, Qt.PenStyle.SolidLine)
        pen.setCapStyle(Qt.PenCapStyle.RoundCap)
        self._painter.setPen(pen)
        self._painter.drawLine(int(x1), int(y1), int(x2), int(y2))

    # Bounding rect for a unit circle, used by _draw_principal_arcs with QTransform
    _UNIT_RECT = QRectF(-1.0, -1.0, 2.0, 2.0)

    def _draw_principal_arcs(self, atom: Atom, r1: float, r2: float, angle: float) -> None:
        """Draw ORTEP-style curved arcs for the three principal planes of the ADP ellipsoid."""
        if getattr(atom, 'u_eigvals', None) is None or np.any(atom.u_eigvals <= 0):
            self._painter.drawLine(int(-r1), 0, int(r1), 0)
            self._painter.drawLine(0, int(-r2), 0, int(r2))
            return

        eigenvalues = atom.u_eigvals
        eigenvectors = atom.u_eigvecs

        angle_rad = radians(angle)
        cos_a = cos(angle_rad)
        sin_a = sin(angle_rad)
        c = self.adp_scale
        s = self.scale

        # Use cosmetic pen so line width is independent of the QTransform scale
        pen = self._painter.pen()
        pen.setCosmetic(True)
        pen.setWidthF(self._adp_intersection_line_width())
        self._painter.setPen(pen)
        self._painter.setBrush(Qt.BrushStyle.NoBrush)

        base_transform = self._painter.transform()

        # Each principal plane is normal to one eigenvector and spanned by the other two
        plane_pairs = ((1, 2), (0, 2), (0, 1))

        for i_ax, j_ax in plane_pairs:
            li = eigenvalues[i_ax]
            lj = eigenvalues[j_ax]
            if li <= 0 or lj <= 0:
                continue

            ri_3d = c * sqrt(li)
            rj_3d = c * sqrt(lj)

            vi = eigenvectors[:, i_ax]
            vj = eigenvectors[:, j_ax]

            Ax = s * ri_3d * (vi[0] * cos_a + vi[1] * sin_a)
            Bx = s * rj_3d * (vj[0] * cos_a + vj[1] * sin_a)
            Ay = s * ri_3d * (-vi[0] * sin_a + vi[1] * cos_a)
            By = s * rj_3d * (-vj[0] * sin_a + vj[1] * cos_a)

            arc_xform = QTransform(Ax, Ay, Bx, By, 0.0, 0.0)
            self._painter.setTransform(arc_xform * base_transform)

            Az = ri_3d * vi[2]
            Bz = rj_3d * vj[2]
            z_amp = sqrt(Az * Az + Bz * Bz)

            if z_amp < 1e-8:
                self._painter.drawArc(self._UNIT_RECT, 0, 5760)
            else:
                phi_z = atan2(Bz, Az)
                start_deg = degrees(-(phi_z + 1.5 * pi))
                self._painter.drawArc(self._UNIT_RECT, int(start_deg * 16), 2880)

        # Restore the original painter transform
        self._painter.setTransform(base_transform)

    def _draw_selection(self, r1: float, r2: float):
        """Draws a crisp, thick cyan outline to indicate selection."""
        padding = 4.0  # pixels
        pen = QPen(QColor(0, 190, 255), max(3, 12 * self._factor), Qt.PenStyle.SolidLine)
        pen.setJoinStyle(Qt.PenJoinStyle.RoundJoin)
        self._painter.setPen(pen)
        self._painter.setBrush(Qt.BrushStyle.NoBrush)

        self._painter.drawEllipse(QRectF(-r1 - padding, -r2 - padding, (r1 + padding) * 2, (r2 + padding) * 2))

    def draw_atom(self, atom: Atom) -> None:
        """Draw a single atom as an ADP ellipsoid, isotropic sphere, or fixed-size circle."""
        if self._show_adps and atom.u_cart is not None:
            if not atom.adp_valid:
                self._draw_invalid_adp(atom)
                return
        cx = atom.screenx
        cy = atom.screeny

        if self._show_adps and atom.u_cart is not None:
            a = atom.u_cart[0, 0]
            b = atom.u_cart[0, 1]
            c = atom.u_cart[1, 1]

            T = a + c
            D = a * c - b * b
            diff = T * T * 0.25 - D

            if diff >= 0:
                sq = sqrt(diff)
                eig1 = T * 0.5 - sq
                eig2 = T * 0.5 + sq

                if eig1 > 0 and eig2 > 0:
                    r1 = sqrt(eig1) * self.scale * self.adp_scale
                    r2 = sqrt(eig2) * self.scale * self.adp_scale

                    if abs(b) > 1e-8:
                        angle = degrees(atan2(eig1 - a, b))
                    else:
                        angle = 0.0 if a < c else 90.0

                    self._painter.save()
                    self._painter.translate(cx, cy)
                    self._painter.rotate(angle)

                    # Draw selection outline matching the ellipsoid shape
                    if atom.name in self.selected_atoms:
                        self._draw_selection(r1, r2)

                    max_r = max(r1, r2)
                    sx, sy = -max_r * 0.3, -max_r * 0.3
                    rad = radians(angle)
                    fx = sx * cos(rad) + sy * sin(rad)
                    fy = -sx * sin(rad) + sy * cos(rad)

                    t = QTransform()
                    t.translate(fx, fy)
                    t.scale(max_r * 1.5, max_r * 1.5)
                    atom.adp_brush.setTransform(t)

                    pen = QPen(self.adp_pen_color, 1, Qt.PenStyle.SolidLine)
                    self._painter.setBrush(atom.adp_brush)
                    self._painter.setPen(pen)
                    self._painter.drawEllipse(QRectF(-r1, -r2, 2 * r1, 2 * r2))

                    cross_pen = QPen(QColor(0, 0, 0, 120), self._adp_intersection_line_width(),
                                     Qt.PenStyle.SolidLine)
                    cross_pen.setCosmetic(True)
                    self._painter.setPen(cross_pen)
                    self._draw_principal_arcs(atom, r1, r2, angle)

                    self._painter.restore()
                    return

        circle_size = self.atoms_size
        if self._show_adps and atom.u_iso is not None:
            r = sqrt(atom.u_iso) * self.scale * self.adp_scale
            circle_size = r * 2

        radius = circle_size / 2

        self._painter.save()
        self._painter.translate(cx, cy)

        # Draw selection outline for spheres
        if atom.name in self.selected_atoms:
            self._draw_selection(radius, radius)

        self._painter.setPen(QPen(self.fallback_pen_color, 1, Qt.PenStyle.SolidLine))
        self._painter.setBrush(atom.sphere_brush)
        self._painter.drawEllipse(QRectF(-radius, -radius, circle_size, circle_size))

        self._painter.restore()

    def _draw_invalid_adp(self, atom: Atom) -> None:
        cx = atom.screenx
        cy = atom.screeny

        size = self.atoms_size

        self._painter.save()
        self._painter.translate(cx, cy)

        # Selection highlight
        if atom.name in self.selected_atoms:
            self._draw_selection(size / 2, size / 2)

        # Cube points (isometric projection)
        s = size * 0.4
        dx = size * 0.3
        dy = -size * 0.3

        # Front face
        fl = QtCore.QPointF(-s - dx / 2, s - dy / 2)
        fr = QtCore.QPointF(s - dx / 2, s - dy / 2)
        tl = QtCore.QPointF(-s - dx / 2, -s - dy / 2)
        tr = QtCore.QPointF(s - dx / 2, -s - dy / 2)

        # Back face visible corners
        btl = QtCore.QPointF(-s + dx / 2, -s + dy / 2)
        btr = QtCore.QPointF(s + dx / 2, -s + dy / 2)
        bbr = QtCore.QPointF(s + dx / 2, s + dy / 2)

        # Faces
        front_face = [tl, tr, fr, fl]
        top_face = [tl, btl, btr, tr]
        right_face = [tr, btr, bbr, fr]

        # Colors for shading (light source top-left)
        color_base = atom.color
        color_light = atom.color_light
        color_dark = atom.color_dark

        pen = QPen(self.fallback_pen_color, 1)
        self._painter.setPen(pen)

        # Draw faces with simple shading
        self._painter.setBrush(QBrush(color_light))
        self._painter.drawPolygon(QtGui.QPolygonF(top_face))
        self._painter.setBrush(QBrush(color_base))
        self._painter.drawPolygon(QtGui.QPolygonF(front_face))
        self._painter.setBrush(QBrush(color_dark))
        self._painter.drawPolygon(QtGui.QPolygonF(right_face))
        # self.draw_npd_text(dx, dy, s)
        self._painter.restore()

    def draw_npd_text(self, dx: float, dy: float, s: float):
        self._painter.setPen(QPen(Qt.GlobalColor.white))
        font = self._painter.font()
        old_size = font.pixelSize()
        font.setPixelSize(max(int(self.atoms_size * 0.3), 1))
        self._painter.setFont(font)
        front_rect = QRectF(-s - dx / 2, -s - dy / 2, 2 * s, 2 * s)
        self._painter.drawText(front_rect, Qt.AlignmentFlag.AlignCenter, "NPD")
        font.setPixelSize(old_size)
        self._painter.setFont(font)

    def draw_label(self, atom: Atom):
        """Draw the atom's name next to its ellipsoid/sphere."""
        self._painter.setPen(QPen(QColor(100, 50, 5), 2, Qt.PenStyle.SolidLine))
        r_pix = self.get_spherical_radius(atom) * self.scale
        self._painter.drawText(int(atom.screenx + r_pix + 2), int(atom.screeny - r_pix - 2), atom.name)

    def get_conntable_from_atoms(self, extra_param: float = 1.2) -> tuple:
        """Build a connectivity table from atomic coordinates and covalent radii."""
        connections = []
        h = ('H', 'D')
        for num1, at1 in enumerate(self.atoms, 0):
            for num2, at2 in enumerate(self.atoms, 0):
                if num1 == num2:
                    continue
                if (at1.part != 0 and at2.part != 0) and at1.part != at2.part:
                    continue
                d = dist(at1.coordinate, at2.coordinate)
                if d > 4.0:
                    continue
                if (at1.radius + at2.radius) * extra_param > d:
                    if at1.type_ in h and at2.type_ in h:
                        continue
                    if (num2, num1) in connections:
                        continue
                    connections.append((num1, num2))
        return tuple(connections)


class Atom:
    """Internal representation of a single atom for rendering."""

    __slots__ = ['coordinate', 'name', 'part', 'radius', 'screenx', 'screeny', 'type_', 'u_cart', 'color',
                 'color_light', 'color_dark', 'u_iso', 'z', 'u_eigvals', 'u_eigvecs', 'u_inv',
                 'sphere_brush', 'adp_brush', 'adp_valid']

    def __init__(self, x: float, y: float, z: float, name: str, type_: str, part: int) -> None:
        self.coordinate = np.array([x, y, z], dtype=np.float32)
        self.adp_valid = True
        self.z = z
        self.name = name
        self.part = part
        self.type_ = type_
        self.screenx = 0
        self.screeny = 0
        self.radius = get_radius_from_element(type_)
        self.u_cart = None
        self.color = QColor(element2color.get(self.type_, '#000000'))  # Default to black if unknown
        self.color_light = self.color.lighter(160)
        self.color_dark = self.color.darker(180)
        self.u_iso = None
        self.u_eigvals = None
        self.u_eigvecs = None
        self.u_inv = None

        self.sphere_brush = QBrush()
        sg = QRadialGradient(0.35, 0.35, 1.0)
        sg.setCoordinateMode(QRadialGradient.CoordinateMode.ObjectBoundingMode)
        sg.setColorAt(0.0, self.color_light)
        sg.setColorAt(0.4, self.color)
        sg.setColorAt(1.0, self.color_dark)
        self.sphere_brush = QBrush(sg)

        ag = QRadialGradient(0.0, 0.0, 1.0)
        ag.setColorAt(0.0, self.color_light)
        ag.setColorAt(0.4, self.color)
        ag.setColorAt(1.0, self.color_dark)
        self.adp_brush = QBrush(ag)

    def __repr__(self) -> str:
        return str((self.name, self.type_, self.coordinate))



"""Tests for :mod:`fastmolwidget.loader`."""

from pathlib import Path

import pytest
from qtpy import QtWidgets

from fastmolwidget.loader import MoleculeLoader
from fastmolwidget.molecule2D import MoleculeWidget

app = QtWidgets.QApplication.instance()
if not app:
    app = QtWidgets.QApplication([])

data = Path('tests/test-data')


# ------------------------------------------------------------------
# Fixtures
# ------------------------------------------------------------------

@pytest.fixture()
def widget():
    w = MoleculeWidget()
    w.resize(400, 300)
    return w


@pytest.fixture()
def loader(widget):
    return MoleculeLoader(widget)


# ------------------------------------------------------------------
# Basic construction
# ------------------------------------------------------------------

def test_loader_holds_widget(widget):
    loader = MoleculeLoader(widget)
    assert loader.widget is widget


# ------------------------------------------------------------------
# CIF loading
# ------------------------------------------------------------------

def test_load_cif(loader, widget):
    loader.load_file(data / '1979688_small.cif')
    assert len(widget.atoms) == 94


def test_load_cif_str_path(loader, widget):
    loader.load_file(str(data / '1979688_small.cif'))
    assert len(widget.atoms) == 94


# ------------------------------------------------------------------
# SHELX loading
# ------------------------------------------------------------------

def test_load_shelx_res(loader, widget):
    loader.load_file(data / 'test_molecule.res')
    assert len(widget.atoms) == 5
    labels = [a.name for a in widget.atoms]
    assert 'C1' in labels
    assert 'O1' in labels


def test_parse_shelx_cell(loader):
    atoms, cell, adps = MoleculeLoader._parse_shelx(data / 'test_molecule.res')
    assert len(cell) == 6
    assert cell[0] == pytest.approx(6.0)
    assert cell[3] == pytest.approx(90.0)


def test_parse_shelx_types(loader):
    atoms, _, _ = MoleculeLoader._parse_shelx(data / 'test_molecule.res')
    types = [a.type for a in atoms]
    assert 'C' in types
    assert 'O' in types
    assert 'H' in types


def test_load_shelx_res_disordered(loader, widget):
    """Test loading a .res file with disorder (multiple PART instructions)."""
    loader.load_file(data / 'p31c-finalcif.res')
    # Should only include PART 0 and PART 1 atoms, not PART 2
    assert len(widget.atoms) > 0
    # Verify no duplicate labels (would happen if both PART 1 and PART 2 were included)
    labels = [a.name for a in widget.atoms]
    assert len(labels) == len(set(labels)), "Duplicate atom labels found - disorder parts not filtered"
    # Should have around 50-60 atoms (PART 0 and PART 1 only, no PART 2)
    assert len(widget.atoms) < 100


def test_parse_shelx_excludes_q_peaks():
    """Test that Q-peaks (residual electron density) are not included as atoms."""
    atoms, _, _ = MoleculeLoader._parse_shelx(data / 'p31c-finalcif.res')
    q_labels = [a.label for a in atoms if a.label.startswith('Q')]
    assert len(q_labels) == 0, f"Q-peaks should be excluded, found: {q_labels}"


def test_parse_shelx_returns_adps():
    """Test that anisotropic displacement parameters are extracted from .res files."""
    atoms, cell, adps = MoleculeLoader._parse_shelx(data / 'p31c-finalcif.res')
    # Should have ADPs for anisotropic atoms (non-hydrogen, non-isotropic)
    assert len(adps) > 0, "Should have ADPs for anisotropic atoms"
    # CL1 should have ADPs
    assert 'CL1' in adps, "CL1 should have ADP values"
    u11, u22, u33, u23, u13, u12 = adps['CL1']
    assert u11 == pytest.approx(0.01547)
    assert u22 == pytest.approx(0.01791)
    assert u33 == pytest.approx(0.02428)
    # Hydrogen atoms should NOT have ADPs (they are isotropic/riding)
    h_in_adps = [k for k in adps if k.startswith('H')]
    assert len(h_in_adps) == 0, f"Hydrogen atoms should not have ADPs, found: {h_in_adps}"


# ------------------------------------------------------------------
# XYZ loading
# ------------------------------------------------------------------

def test_load_xyz(loader, widget):
    loader.load_file(data / 'test_molecule.xyz')
    assert len(widget.atoms) == 5


def test_parse_xyz_labels():
    atoms = MoleculeLoader._parse_xyz(data / 'test_molecule.xyz')
    # Labels should be Element + 1-based index
    assert atoms[0].label == 'O1'
    assert atoms[3].label == 'C4'


def test_parse_xyz_coords():
    atoms = MoleculeLoader._parse_xyz(data / 'test_molecule.xyz')
    assert atoms[0].x == pytest.approx(0.0)
    assert atoms[0].z == pytest.approx(0.1173)


# ------------------------------------------------------------------
# Error handling
# ------------------------------------------------------------------

def test_unsupported_format(loader):
    with pytest.raises(ValueError, match='Unsupported file format'):
        loader.load_file(data / 'fake.pdb')


def test_file_not_found(loader):
    with pytest.raises(FileNotFoundError):
        loader.load_file('nonexistent.cif')


def test_xyz_bad_atom_count(tmp_path):
    bad_xyz = tmp_path / 'bad.xyz'
    bad_xyz.write_text('10\ncomment\nC 0 0 0\n')
    with pytest.raises(ValueError, match='declares 10 atoms but 1'):
        MoleculeLoader._parse_xyz(bad_xyz)


def test_xyz_too_short(tmp_path):
    bad_xyz = tmp_path / 'short.xyz'
    bad_xyz.write_text('1\n')
    with pytest.raises(ValueError, match='too short'):
        MoleculeLoader._parse_xyz(bad_xyz)


def test_shelx_no_cell(tmp_path):
    bad_res = tmp_path / 'nocell.res'
    bad_res.write_text('TITL test\nSFAC C\nC1 1 0.1 0.2 0.3 11.0 0.02\nEND\n')
    with pytest.raises(ValueError, match='No CELL instruction'):
        MoleculeLoader._parse_shelx(bad_res)


# ------------------------------------------------------------------
# Keep view
# ------------------------------------------------------------------

def test_load_keep_view(loader, widget):
    loader.load_file(data / 'test_molecule.xyz')
    first_count = len(widget.atoms)
    assert first_count == 5
    # Load again with keep_view
    loader.load_file(data / 'test_molecule.xyz', keep_view=True)
    assert len(widget.atoms) == 5


# ------------------------------------------------------------------
# Rendering sanity check
# ------------------------------------------------------------------

def test_loaded_molecule_renders(loader, widget):
    loader.load_file(data / '1979688_small.cif')
    widget.show()
    pixmap = widget.grab()
    assert not pixmap.isNull()


# ------------------------------------------------------------------
# Grow (SDM)
# ------------------------------------------------------------------

def test_grow_produces_more_atoms_than_asym_unit(widget):
    loader = MoleculeLoader(widget)
    loader.load_file(data / '1979688_small.cif')
    asym_count = len(widget.atoms)
    loader.set_grow(True)
    grown_count = len(widget.atoms)
    assert grown_count >= asym_count, "Grown structure should have at least as many atoms"


def test_grow_toggle_restores_asym_unit(widget):
    loader = MoleculeLoader(widget)
    loader.load_file(data / '1979688_small.cif')
    asym_count = len(widget.atoms)
    loader.set_grow(True)
    loader.set_grow(False)
    assert len(widget.atoms) == asym_count


def test_grow_has_no_effect_on_xyz(widget):
    loader = MoleculeLoader(widget)
    loader.load_file(data / 'test_molecule.xyz')
    count_before = len(widget.atoms)
    loader.set_grow(True)  # should silently do nothing for non-CIF
    assert len(widget.atoms) == count_before


def test_grow_shelx_produces_more_atoms(widget):
    """Grow should expand the asymmetric unit for .res files too."""
    loader = MoleculeLoader(widget)
    loader.load_file(data / 'p31c-finalcif.res')
    asym_count = len(widget.atoms)
    loader.set_grow(True)
    grown_count = len(widget.atoms)
    assert grown_count > asym_count, "Grown .res structure should have more atoms"


def test_grow_shelx_toggle_restores_asym_unit(widget):
    """Toggling grow off should restore the original atom count for .res files."""
    loader = MoleculeLoader(widget)
    loader.load_file(data / 'p31c-finalcif.res')
    asym_count = len(widget.atoms)
    loader.set_grow(True)
    loader.set_grow(False)
    assert len(widget.atoms) == asym_count





from __future__ import annotations

import itertools


def to_float(st) -> float | list[float] | None:
    if isinstance(st, list):
        try:
            return [float(x) for x in st[-2:]]
        except ValueError:
            return None
    else:
        try:
            return float(st.split('(')[0])
        except ValueError:
            return None


def get_error_from_value(value: str) -> tuple[float, float]:
    """
    Returns the error value from a number string.
    """
    try:
        value = value.replace(" ", "")
    except AttributeError:
        return float(value), 0.0
    if "(" in value:
        vval, err = value.split("(")
        val = vval.split('.')
        err = err.split(")")[0]
        if not err:
            return float(vval), 0.0
        if len(val) > 1:
            return float(vval), int(err) * (10 ** (-1 * len(val[1])))
        else:
            return float(vval), float(err)
    else:
        try:
            return float(value), 0.0
        except ValueError:
            return 0.0, 0.0


def isnumeric(value: str) -> bool:
    """
    Determines if a string can be converted to a number.
    """
    value = value.split('(')[0]
    try:
        float(value)
    except ValueError:
        return False
    return True


def grouper(inputs, n, fillvalue=None):
    iters = [iter(inputs)] * n
    return itertools.zip_longest(*iters, fillvalue=fillvalue)



# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#


import time
from collections import namedtuple
from math import sqrt, cos, radians, sin, floor
from typing import TYPE_CHECKING

try:
    import sdm_cpp

    HAS_CPP = True
except ImportError:
    HAS_CPP = False

from fastmolwidget.atoms import get_radius_from_element
from fastmolwidget.dsrmath import Array, SymmetryElement, Matrix, frac_to_cart

if TYPE_CHECKING:
    from cif.cif_file_io import CifReader

DEBUG = False
Atomtuple = namedtuple('Atomtuple', ('label', 'type', 'x', 'y', 'z', 'part', 'symm_matrix'), defaults=(None,))


class SymmCards:
    """
    Contains the list of SYMM cards
    """

    def __init__(self):
        self._symmcards = [SymmetryElement(['X', 'Y', 'Z'])]

    def _as_str(self) -> str:
        return "\n".join([str(x) for x in self._symmcards])

    def __repr__(self) -> str:
        return self._as_str()

    def __str__(self) -> str:
        return self._as_str()

    def __getitem__(self, item):
        return self._symmcards[item]

    def __iter__(self):
        yield from self._symmcards

    def __len__(self):
        return len(self._symmcards)

    def append(self, symmData: list) -> None:
        """
        Add the content of a Shelxl SYMM command to generate the appropriate SymmetryElement instance.
        :param symmData: list of strings. eg.['1/2+X', '1/2+Y', '1/2+Z']
        :return: None
        """
        newSymm = SymmetryElement(symmData)
        if newSymm not in self._symmcards:
            self._symmcards.append(newSymm)


class SDMItem:
    __slots__ = ['a1', 'a2', 'atom1', 'atom2', 'covalent', 'dddd', 'dist', 'symmetry_number']

    def __init__(self):
        self.dist = 0.0
        self.atom1 = None
        self.a1 = 0
        self.atom2 = None
        self.a2 = 0
        self.symmetry_number = 0
        self.covalent = True
        self.dddd = 0

    def __lt__(self, a2) -> bool:
        return True if self.dist < a2.dist else False

    def __eq__(self, other: 'SDMItem') -> bool:
        if other.a1 == self.a2 and other.a2 == self.a1:
            return True
        return False

    def __repr__(self):
        return f'{self.atom1.name} {self.atom2.name} {self.a1} {self.a2} dist: {self.dist} coval: {self.covalent} sn: {self.symmetry_number} {self.dddd}'


class SDM:
    sdm_list: list[SDMItem]

    def __init__(self, atoms: tuple[list], symmlist: list, cell: tuple[float, float, float, float, float, float],
                 centric=False):
        """
        Calculates the shortest distance matrix
                        0      1      2  3  4   5     6          7
        :param atoms: [Name, Element, X, Y, Z, Part, ocuupancy, molindex -> (later)]
        :param symmlist:
        :param cell:
        """
        self.atoms = atoms
        self.symmcards = SymmCards()
        if centric:
            self.symmcards.append(['-X', '-Y', '-Z'])
            self.symmcards[-1].centric = True
        for s in symmlist:
            self.symmcards.append(s.split(','))
        self.cell = cell
        self.cosal = cos(radians(cell[3]))
        self.cosbe = cos(radians(cell[4]))
        self.cosga = cos(radians(cell[5]))
        self.aga = self.cell[0] * self.cell[1] * self.cosga
        self.bbe = self.cell[0] * self.cell[2] * self.cosbe
        self.cal = self.cell[1] * self.cell[2] * self.cosal
        self.asq = self.cell[0] ** 2
        self.bsq = self.cell[1] ** 2
        self.csq = self.cell[2] ** 2
        self.sdm_list = []  # list of sdmitems
        self.maxmol = 1
        self.sdmtime = 0

    def orthogonal_matrix(self) -> Matrix:
        """
        Converts von fractional to cartesian.
        """
        return Matrix([[self.cell[0], self.cell[1] * cos(self.cell[5]), self.cell[2] * cos(self.cell[4])],
                       [0, self.cell[1] * sin(self.cell[5]),
                        (self.cell[2] * (cos(self.cell[3]) - cos(self.cell[4]) * cos(self.cell[5])) / sin(
                            self.cell[5]))],
                       [0, 0, self.cell[6] / (self.cell[0] * self.cell[1] * sin(self.cell[5]))]])

    def calc_sdm(self) -> list:
        t1 = time.perf_counter()
        h = {'H', 'D'}
        nlen = len(self.symmcards)

        symm_m = []
        symm_t = []
        for s in self.symmcards:
            symm_m.append(tuple(tuple(row) for row in s.matrix.values))
            symm_t.append(tuple(s.trans))

        # C++ Fast Path
        if HAS_CPP:
            coords = [[at[2], at[3], at[4]] for at in self.atoms]
            radii = [get_radius_from_element(at[1]) for at in self.atoms]
            is_h = [at[1] in h for at in self.atoms]
            parts = [at[5] for at in self.atoms]

            cpp_results = sdm_cpp.calc_sdm_cpp(
                coords, symm_m, symm_t,
                self.aga, self.bbe, self.cal,
                self.asq, self.bsq, self.csq,
                radii, is_h, parts
            )

            for (i, j, best_n, mind, dddd, covalent) in cpp_results:
                sdm_item = SDMItem()
                sdm_item.dist = mind
                sdm_item.atom1 = self.atoms[i]
                sdm_item.atom2 = self.atoms[j]
                sdm_item.a1 = i
                sdm_item.a2 = j
                sdm_item.symmetry_number = best_n
                sdm_item.dddd = dddd
                sdm_item.covalent = covalent
                self.sdm_list.append(sdm_item)

        # Pure Python Fallback Path
        else:
            at2_plushalf = [(x[2] + 0.5, x[3] + 0.5, x[4] + 0.5) for x in self.atoms]
            aga, bbe, cal = self.aga, self.bbe, self.cal
            asq, bsq, csq = self.asq, self.bsq, self.csq

            for i, at1 in enumerate(self.atoms):
                x1, y1, z1 = at1[2], at1[3], at1[4]

                prime_array = []
                for m, t in zip(symm_m, symm_t):
                    px = x1 * m[0][0] + y1 * m[1][0] + z1 * m[2][0] + t[0]
                    py = x1 * m[0][1] + y1 * m[1][1] + z1 * m[2][1] + t[1]
                    pz = x1 * m[0][2] + y1 * m[1][2] + z1 * m[2][2] + t[2]
                    prime_array.append((px, py, pz))

                for j, at2 in enumerate(self.atoms):
                    mind = 1000000.0
                    hma = False
                    atp_x, atp_y, atp_z = at2_plushalf[j]

                    sdm_item = SDMItem()

                    for n in range(nlen):
                        px, py, pz = prime_array[n]

                        dx = px - atp_x
                        dy = py - atp_y
                        dz = pz - atp_z

                        dpx = dx - floor(dx) - 0.5
                        dpy = dy - floor(dy) - 0.5
                        dpz = dz - floor(dz) - 0.5

                        A = 2.0 * (dpx * dpy * aga + dpx * dpz * bbe + dpy * dpz * cal)
                        dk2 = dpx * dpx * asq + dpy * dpy * bsq + dpz * dpz * csq + A

                        if dk2 > 16.0:
                            continue

                        dk = sqrt(dk2)
                        if n:
                            dk += 0.0001

                        if (dk > 0.01) and (mind >= dk):
                            mind = dk
                            sdm_item.dist = mind
                            sdm_item.atom1 = at1
                            sdm_item.atom2 = at2
                            sdm_item.a1 = i
                            sdm_item.a2 = j
                            sdm_item.symmetry_number = n
                            hma = True

                    if not sdm_item.atom1:
                        continue
                    if ((sdm_item.atom1[1] not in h and sdm_item.atom2[1] not in h) and
                        sdm_item.atom1[5] * sdm_item.atom2[5] == 0) or sdm_item.atom1[5] == sdm_item.atom2[5]:
                        dddd = (get_radius_from_element(at1[1]) + get_radius_from_element(at2[1])) * 1.2
                        sdm_item.dddd = dddd
                    else:
                        dddd = 0.0
                    if sdm_item.dist < dddd:
                        if hma:
                            sdm_item.covalent = True
                    else:
                        sdm_item.covalent = False
                    if hma:
                        self.sdm_list.append(sdm_item)

        t2 = time.perf_counter()
        self.sdmtime = t2 - t1
        print(f'Time for sdm {"(C++)" if HAS_CPP else "(Python fallback)"}:', round(self.sdmtime, 4), 's')

        self.sdm_list.sort()
        self.calc_molindex(self.atoms)
        need_symm = self.collect_needed_symmetry()
        if DEBUG:
            print(f"The asymmetric unit contains {self.maxmol} fragments.")
        return need_symm

    def collect_needed_symmetry(self) -> list:
        need_symm = []
        h = ('H', 'D')

        symm_m = []
        symm_t = []
        for s in self.symmcards:
            symm_m.append(tuple(tuple(row) for row in s.matrix.values))
            symm_t.append(tuple(s.trans))

        aga, bbe, cal = self.aga, self.bbe, self.cal
        asq, bsq, csq = self.asq, self.bsq, self.csq

        # Collect needsymm list:
        for sdm_item in self.sdm_list:
            if sdm_item.covalent:
                if sdm_item.atom1[-1] < 1 or sdm_item.atom1[-1] > 6:
                    continue

                x1, y1, z1 = sdm_item.atom1[2], sdm_item.atom1[3], sdm_item.atom1[4]
                x2, y2, z2 = sdm_item.atom2[2], sdm_item.atom2[3], sdm_item.atom2[4]

                for n, (m, t) in enumerate(zip(symm_m, symm_t)):
                    if sdm_item.atom1[5] * sdm_item.atom2[5] != 0 and \
                            sdm_item.atom1[5] != sdm_item.atom2[5]:
                        continue
                    # Both the same atomic number and number 0 (hydrogen)
                    if sdm_item.atom1[1] == sdm_item.atom2[1] and sdm_item.atom1[1] in h:
                        continue

                    px = x1 * m[0][0] + y1 * m[1][0] + z1 * m[2][0] + t[0]
                    py = x1 * m[0][1] + y1 * m[1][1] + z1 * m[2][1] + t[1]
                    pz = x1 * m[0][2] + y1 * m[1][2] + z1 * m[2][2] + t[2]

                    Dx = px - x2 + 0.5
                    Dy = py - y2 + 0.5
                    Dz = pz - z2 + 0.5

                    fDx, fDy, fDz = floor(Dx), floor(Dy), floor(Dz)

                    dpx = Dx - fDx - 0.5
                    dpy = Dy - fDy - 0.5
                    dpz = Dz - fDz - 0.5

                    if n == 0 and fDx == 0 and fDy == 0 and fDz == 0:
                        continue

                    A = 2.0 * (dpx * dpy * aga + dpx * dpz * bbe + dpy * dpz * cal)
                    dk2 = dpx * dpx * asq + dpy * dpy * bsq + dpz * dpz * csq + A

                    if dk2 <= 0.000001:
                        continue

                    dk = sqrt(dk2)
                    dddd = sdm_item.dist + 0.2
                    if sdm_item.atom1[1] in h and sdm_item.atom2[1] in h:
                        dddd = 1.8

                    if dk <= dddd:
                        bs = [n + 1, int(5 - fDx), int(5 - fDy), int(5 - fDz), sdm_item.atom1[-1]]
                        if bs not in need_symm:
                            need_symm.append(bs)
        return need_symm

    def calc_molindex(self, all_atoms):
        # Start for George's "bring atoms together algorithm":
        someleft = 1
        nextmol = 1
        for at in all_atoms:
            at.append(-1)
        all_atoms[0][-1] = 1
        while nextmol:
            someleft = 1
            nextmol = 0
            while someleft:
                someleft = 0
                for sdm_item in self.sdm_list:
                    if sdm_item.covalent and sdm_item.atom1[-1] * sdm_item.atom2[-1] < 0:
                        sdm_item.atom1[-1] = self.maxmol  # last item is the molindex
                        sdm_item.atom2[-1] = self.maxmol
                        someleft += 1
            for ni, at in enumerate(all_atoms):
                if at[-1] < 0:
                    nextmol = ni
                    break
            if nextmol:
                self.maxmol += 1
                all_atoms[nextmol][-1] = self.maxmol

    def vector_length(self, x: float, y: float, z: float) -> float:
        """
        Calculates the vector length given in fractional coordinates.
        """
        A = 2.0 * (x * y * self.aga + x * z * self.bbe + y * z * self.cal)
        return sqrt(x ** 2 * self.asq + y ** 2 * self.bsq + z ** 2 * self.csq + A)

    def packer(self, sdm: 'SDM', need_symm: list, with_qpeaks=False) -> list[Atomtuple]:
        """
        Packs atoms of the asymmetric unit to real molecules.
        """
        showatoms = []
        # Append base atoms with an identity matrix
        identity_matrix = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0))
        for at in self.atoms:
            showatoms.append(list(at) + ['base', identity_matrix])

        new_atoms = []

        symm_m = []
        symm_t = []
        for s in self.symmcards:
            symm_m.append(tuple(tuple(row) for row in s.matrix.values))
            symm_t.append(tuple(s.trans))

        for symm in need_symm:
            s, h, k, l, symmgroup = symm
            h -= 5
            k -= 5
            l -= 5
            s -= 1

            m = symm_m[s]
            t = symm_t[s]

            for atom in self.atoms:
                if atom[-1] == symmgroup:
                    x1, y1, z1 = atom[2], atom[3], atom[4]
                    px = x1 * m[0][0] + y1 * m[1][0] + z1 * m[2][0] + t[0] + h
                    py = x1 * m[0][1] + y1 * m[1][1] + z1 * m[2][1] + t[1] + k
                    pz = x1 * m[0][2] + y1 * m[1][2] + z1 * m[2][2] + t[2] + l

                    # The new atom with the symmetry matrix appended:
                    new = [atom[0], atom[1], px, py, pz, atom[5], atom[6], atom[7], 'symmgen', m]
                    new_atoms.append(new)
                    isthere = False
                    # Only add atom if its occupancy (new[5]) is greater zero:
                    if new[5] >= 0:
                        for existing in showatoms:
                            if existing[5] != new[5]:
                                continue
                            length = sdm.vector_length(px - existing[2],
                                                       py - existing[3],
                                                       pz - existing[4])
                            if length < 0.2:
                                isthere = True
                                break
                    if not isthere:
                        showatoms.append(new)

        cart_atoms = []
        cell = self.cell[:6]
        for at in showatoms:
            x, y, z = frac_to_cart([at[2], at[3], at[4]], cell)
            # at[-1] contains the symm_matrix
            cart_atoms.append(Atomtuple(label=at[0], type=at[1], x=x, y=y, z=z, part=at[5], symm_matrix=at[-1]))
        return cart_atoms


if __name__ == "__main__":
    import sys
    from pathlib import Path
    from qtpy import QtWidgets
    from fastmolwidget.viewer_widget import MoleculeViewerWidget

    app = QtWidgets.QApplication(sys.argv)
    viewer = MoleculeViewerWidget(Path('tests/test-data/4060314.cif'))
    viewer.show()
    sys.exit(app.exec())



"""
Loader class for :class:`~fastmolwidget.molecule2D.MoleculeWidget`.

Uses composition to provide file-format–aware loading of molecular data into a
:class:`MoleculeWidget`.  Supported formats:

* **CIF** – Crystallographic Information File (``.cif``)
* **SHELX** – SHELXL instruction file (``.res`` / ``.ins``)
* **XYZ** – Standard XYZ coordinate file (``.xyz``)

Example usage::

    widget = MoleculeWidget()
    loader = MoleculeLoader(widget)
    loader.load_file("structure.cif")
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Generator

from shelxfile import Shelxfile

from fastmolwidget.cif.cif_file_io import CifReader, adp
from fastmolwidget.molecule2D import MoleculeWidget
from fastmolwidget.sdm import Atomtuple
from fastmolwidget.tools import to_float


class MoleculeLoader:
    """Load molecular structures from various file formats into a
    :class:`MoleculeWidget`.

    The loader uses **composition**: it holds a reference to the widget and
    delegates rendering to it via :meth:`MoleculeWidget.open_molecule`.

    :param widget: The molecule widget to populate.
    """

    _FORMAT_MAP: dict[str, str] = {
        '.cif': '_load_cif',
        '.res': '_load_shelx',
        '.ins': '_load_shelx',
        '.xyz': '_load_xyz',
    }

    def __init__(self, widget: MoleculeWidget) -> None:
        self._widget = widget
        self._grow_enabled: bool = False
        self._last_path: Path | None = None

    @property
    def widget(self) -> MoleculeWidget:
        """The :class:`MoleculeWidget` this loader populates."""
        return self._widget

    def load_file(self, path: str | Path, *, keep_view: bool = False) -> None:
        """Load a molecular structure from *path*.

        The file format is determined from the file extension.

        :param path: Path to the file.
        :param keep_view: If ``True``, preserve the current zoom / rotation.
        :raises ValueError: If the file format is not supported.
        :raises FileNotFoundError: If the file does not exist.
        """
        path = Path(path)
        self._last_path = path
        suffix = path.suffix.lower()
        loader_name = self._FORMAT_MAP.get(suffix)
        if loader_name is None:
            supported = ', '.join(sorted(self._FORMAT_MAP))
            raise ValueError(
                f"Unsupported file format '{suffix}'. "
                f"Supported extensions: {supported}"
            )
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")
        loader = getattr(self, loader_name)
        loader(path, keep_view=keep_view)

    _GROWABLE_FORMATS: frozenset[str] = frozenset({'.cif', '.res', '.ins'})

    def set_grow(self, enabled: bool) -> None:
        """Toggle SDM grow mode and reload the current file if it supports
        crystal symmetry (CIF or SHELX).

        Grow expands the asymmetric unit to complete molecules using crystal
        symmetry.  Has no effect when the last loaded file is an XYZ (no
        symmetry information).

        :param enabled: ``True`` to enable structure growing, ``False`` to
            revert to the bare asymmetric unit.
        """
        self._grow_enabled = enabled
        if (self._last_path is not None
                and self._last_path.suffix.lower() in self._GROWABLE_FORMATS):
            self.load_file(self._last_path, keep_view=True)

    # ------------------------------------------------------------------
    # CIF loading
    # ------------------------------------------------------------------

    def _load_cif(self, path: Path, *, keep_view: bool = False) -> None:
        """Load a CIF file using :class:`CifReader`."""
        cif = CifReader(path)
        atoms = (
            self._compute_grown_atoms(cif)
            if self._grow_enabled
            else list(cif.atoms_orth)
        )
        self._widget.open_molecule(
            atoms=atoms,
            cell=cif.cell[:6],
            adps=self._load_adps_from_cif(cif.displacement_parameters()),
            keep_view=keep_view,
        )

    @staticmethod
    def _compute_grown_atoms(cif: CifReader) -> list:
        """Expand the asymmetric unit to complete molecules via the SDM.

        Reads fractional-coordinate atoms, runs the Shortest Distance Matrix
        algorithm with the CIF's symmetry operators, and returns the packed
        Cartesian-coordinate atom list.

        :param cif: The parsed CIF to grow.
        :returns: A list of :class:`~fastmolwidget.sdm.Atomtuple` in Cartesian
            coordinates including all symmetry-generated atoms that complete
            the molecule(s).
        """
        from fastmolwidget.sdm import SDM

        # SDM.calc_molindex mutates the atom lists in place – pass a fresh copy
        fract_atoms = list(cif.atoms_fract)
        sdm = SDM(fract_atoms, cif.symmops, cif.cell, centric=cif.is_centrosymm)
        need_symm = sdm.calc_sdm()
        return sdm.packer(sdm, need_symm)

    @staticmethod
    def _load_adps_from_cif(
        adps: Generator[adp, Any, None],
    ) -> dict[str, tuple[float, float, float, float, float, float]]:
        """Convert a generator of CIF displacement parameters into the ADP
        mapping expected by :meth:`MoleculeWidget.open_molecule`.

        :param adps: Generator of :class:`~fastmolwidget.cif.cif_file_io.adp`
            named-tuples produced by
            :meth:`~fastmolwidget.cif.cif_file_io.CifReader.displacement_parameters`.
        :returns: A dict mapping atom labels to ``(U11, U22, U33, U23, U13,
            U12)`` tuples of floats.
        """
        adp_dict: dict[str, tuple[float, float, float, float, float, float]] = {}
        for dp in adps:
            adp_dict[dp.label] = (
                to_float(dp.U11), to_float(dp.U22), to_float(dp.U33),
                to_float(dp.U23), to_float(dp.U13), to_float(dp.U12),
            )
        return adp_dict

    # ------------------------------------------------------------------
    # SHELX .res / .ins loading
    # ------------------------------------------------------------------

    def _load_shelx(self, path: Path, *, keep_view: bool = False) -> None:
        """Load a SHELX instruction (.res / .ins) file using the
        :mod:`shelxfile` library."""
        atoms, cell, adps = self._parse_shelx(path)
        if self._grow_enabled:
            atoms = self._compute_grown_atoms_shelx(path)
        self._widget.open_molecule(atoms=atoms, cell=cell, adps=adps,
                                   keep_view=keep_view)

    @staticmethod
    def _compute_grown_atoms_shelx(path: Path) -> list:
        """Expand the asymmetric unit of a SHELX file to complete molecules
        via the SDM, analogous to :meth:`_compute_grown_atoms` for CIF files.

        :param path: Path to the SHELX ``.res`` / ``.ins`` file.
        :returns: A list of :class:`~fastmolwidget.sdm.Atomtuple` in Cartesian
            coordinates including all symmetry-generated atoms.
        """
        from fastmolwidget.sdm import SDM

        shx = Shelxfile()
        shx.read_file(path)

        cell_params: tuple[float, float, float, float, float, float] = (
            shx.cell.a, shx.cell.b, shx.cell.c,
            shx.cell.alpha, shx.cell.beta, shx.cell.gamma,
        )

        # Build fractional-coordinate atom lists (mutable – SDM mutates them)
        fract_atoms: list[list] = []
        for at in shx.atoms:
            if at.qpeak:
                continue
            x, y, z = at.frac_coords
            fract_atoms.append(
                [at.name, at.element, x, y, z, at.part.n, at.occupancy, at.ueq]
            )

        # Collect symmetry operations as comma-separated strings (skip identity)
        symmops: list[str] = []
        for s in shx.symmcards:
            op_str = s.to_shelxl()
            # SDM already includes identity; only add non-identity ops
            symmops.append(op_str)

        centric = shx.latt.centric if shx.latt else False

        sdm = SDM(fract_atoms, symmops, cell_params, centric=centric)
        need_symm = sdm.calc_sdm()
        return sdm.packer(sdm, need_symm)

    # ------------------------------------------------------------------
    # XYZ loading
    # ------------------------------------------------------------------

    def _load_xyz(self, path: Path, *, keep_view: bool = False) -> None:
        """Load a standard XYZ coordinate file.

        The format is::

            <number of atoms>
            <comment line>
            <element> <x> <y> <z>
            ...

        Coordinates are assumed to be in Ångströms (Cartesian).
        XYZ files have no unit-cell or ADP information.
        """
        atoms = self._parse_xyz(path)
        self._widget.open_molecule(atoms=atoms, cell=None, adps=None,
                                   keep_view=keep_view)

    # ------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------

    @staticmethod
    def _parse_shelx(
        path: Path,
    ) -> tuple[
        list[Atomtuple],
        tuple[float, float, float, float, float, float],
        dict[str, tuple[float, float, float, float, float, float]],
    ]:
        """Parse a SHELX .res / .ins file using the :mod:`shelxfile` library.

        Returns the atom list (in Cartesian coordinates), the unit-cell
        parameters, and a dictionary of anisotropic displacement parameters.

        Q-peaks (residual electron-density peaks) are excluded.
        """
        shx = Shelxfile()
        shx.read_file(path)

        if shx.cell is None:
            raise ValueError(f"No CELL instruction found in SHELX file: {path}")

        cell = shx.cell
        cell_params: tuple[float, float, float, float, float, float] = (
            cell.a, cell.b, cell.c, cell.alpha, cell.beta, cell.gamma,
        )

        atoms: list[Atomtuple] = []
        adp_dict: dict[str, tuple[float, float, float, float, float, float]] = {}
        for at in shx.atoms:
            # Skip Q-peaks (residual electron-density peaks, not real atoms).
            if at.qpeak:
                continue

            x, y, z = at.cart_coords
            atoms.append(Atomtuple(
                label=at.name,
                type=at.element,
                x=x,
                y=y,
                z=z,
                part=at.part.n,
            ))

            # Collect anisotropic displacement parameters for non-isotropic atoms.
            if not at.is_isotropic:
                u11, u22, u33, u23, u13, u12 = at.uvals
                adp_dict[at.name] = (u11, u22, u33, u23, u13, u12)

        return atoms, cell_params, adp_dict

    @staticmethod
    def _parse_xyz(path: Path) -> list[Atomtuple]:
        """Parse a standard XYZ file and return a list of
        :class:`Atomtuple`."""
        lines = path.read_text().splitlines()
        if len(lines) < 3:
            raise ValueError(f"XYZ file too short: {path}")

        try:
            natoms = int(lines[0].strip())
        except ValueError:
            raise ValueError(
                f"First line of XYZ file must be the atom count, "
                f"got: {lines[0].strip()!r}"
            )

        atom_lines = lines[2:]  # skip count and comment
        atoms: list[Atomtuple] = []
        for idx, line in enumerate(atom_lines):
            parts = line.split()
            if len(parts) < 4:
                continue  # skip blank / malformed lines
            element = parts[0]
            try:
                x, y, z = float(parts[1]), float(parts[2]), float(parts[3])
            except ValueError:
                continue
            label = f"{element}{idx + 1}"
            atoms.append(
                Atomtuple(label=label, type=element, x=x, y=y, z=z, part=0)
            )

        if len(atoms) != natoms:
            raise ValueError(
                f"XYZ file declares {natoms} atoms but {len(atoms)} were "
                f"parsed from {path}"
            )
        return atoms
